from . import resources
from .workflow import Workflow, Definition, Job, JobContext
from .configuration import Config

from bigflow._version import __version__


__all__ = [
    # core
    'Workflow',
    'Job',
    'JobContext',
    'Definition',
    'Config',
    'resources'
]

try:
    from . import bigquery
    __all__.append('bigquery')
except ImportError:
    pass

try:
    from . import monitoring
    __all__.append('monitoring')
except ImportError:
    pass



import os
import io
import pprint


class EnvConfig:
    def __init__(self,
                 name: str,
                 properties: dict):
        self.name = name
        self.properties = properties


def current_env():
    """Returns current env name (specified via 'bigflow --config' option)"""
    return os.environ.get('bf_env')


class Config:
    def __init__(self,
                 name: str,
                 properties: dict,
                 is_master: bool = True,
                 is_default: bool = True,
                 environment_variables_prefix: str = 'bf_'):
        if is_master:
            self.master_config_name = name

        self.configs = {
            name: EnvConfig(name, properties)
        }

        self.default_env_name = None
        self._update_default_env_name(name, is_default)

        self.environment_variables_prefix = environment_variables_prefix

        self._check_docker_repository(properties)

    def __str__(self):
        return '\n'.join(list(map(lambda e: self.pretty_print(e), self.configs.keys())))

    def resolve_property(self, property_name: str, env_name: str = None):
        static_props = self.resolve(env_name)

        if property_name in static_props:
            return static_props[property_name]

        return self._resolve_property_from_os_env(property_name)

    def pretty_print(self, env_name: str = None):
        s = io.StringIO()
        pp = pprint.PrettyPrinter(indent=4, stream=s)

        env_config = self._get_env_config(env_name)
        s.write(env_config.name + ' config:\n')
        pp.pprint(self.resolve(env_name))

        return s.getvalue()[:-1]

    def resolve(self, env_name: str = None) -> dict:
        env_config = self._get_env_config(env_name)

        properties_with_placeholders = {key: self._resolve_property_with_os_env_fallback(key, value)
                                        for (key, value) in env_config.properties.items()}

        string_properties = {k: v for (k, v) in properties_with_placeholders.items() if isinstance(v, str)}
        string_properties['env'] = env_config.name

        return {key: self._resolve_placeholders(value, string_properties)
                for (key, value) in properties_with_placeholders.items()}

    def add_configuration(self, name: str, properties: dict, is_default: bool = False):

        all_properties = self._get_master_properties()
        all_properties.update(properties)

        self._check_docker_repository(properties)

        self.configs[name] = EnvConfig(name, all_properties)

        self._update_default_env_name(name, is_default)

        return self

    def _update_default_env_name(self, name: str, is_default: bool):
        if is_default:
            if self.default_env_name:
                raise ValueError(f"default env is already set to '{self.default_env_name}', you can set only one default env")
            self.default_env_name = name

    def _get_master_properties(self) -> dict:
        if not self._has_master_config():
            return {}

        return self._get_env_config(self.master_config_name).properties.copy()

    def _has_master_config(self) -> bool :
        return hasattr(self, 'master_config_name')

    def _get_env_config(self, name: str) -> EnvConfig:

        explicit_env_name = name or self._check_property_from_os_env('env')

        if not explicit_env_name:
            return self._get_default_config()
        else:
            if explicit_env_name not in self.configs:
                raise ValueError(f"no such config name '{explicit_env_name}'")
            return self.configs[explicit_env_name]

    def _get_default_config(self) -> EnvConfig:
        if not self.default_env_name:
            raise ValueError("No explicit env name is given and no default env is defined, can't resolve properties.")
        return self.configs[self.default_env_name]

    def _resolve_property_with_os_env_fallback(self, key, value):
        if value is None:
            return self._resolve_property_from_os_env(key)
        else:
            return value

    def _os_env_variable_name(self, key):
        return self.environment_variables_prefix + key

    def _check_property_from_os_env(self, key):
        env_var_name = self._os_env_variable_name(key)
        if env_var_name in os.environ:
            return os.environ[env_var_name]
        return None

    def _resolve_property_from_os_env(self, property_name):
        property = self._check_property_from_os_env(property_name)
        if not property:
            os_env_var_name = self._os_env_variable_name(property_name)
            raise ValueError(f"Failed to load property '{property_name}' from OS environment, no such env variable: '{os_env_var_name}'.")
        return property

    def _resolve_placeholders(self, value, variables:dict):
        if isinstance(value, str):
            modified_value = value
            for k, v in variables.items():
                if v != value:
                    modified_value = modified_value.replace('{' + k + '}', v)
            return modified_value
        else:
            return value

    def _check_docker_repository(self, properties):
        if "docker_repository" in properties.keys():
            if not properties["docker_repository"].islower():
                raise ValueError("docker_repository variable should be in lower case")


import re
import os
from shutil import rmtree
from uuid import uuid1
import subprocess
from pathlib import Path
import tempfile
from unittest import TestCase, mock

from bigflow.version import VERSION_PATTERN, bump_minor, release, STARTING_VERSION
from bigflow.commons import resolve

NO_REPOSITORY_VERSION_PATTERN = re.compile(r'^0.1.0SNAPSHOT\w+$')
NO_COMMITS_VERSION_PATTERN = NO_REPOSITORY_VERSION_PATTERN

NO_TAG_VERSION_PATTERN = re.compile(r'^0.1.0SNAPSHOT\w+$')
NO_TAG_DIRTY_VERSION_PATTERN = re.compile(r'^0.1.0SNAPSHOT\w+$')

TAG_ON_HEAD_VERSION_PATTERN = re.compile(r'^\d+\.\d+\.\d+$')
TAG_ON_HEAD_DIRTY_VERSION_PATTERN = re.compile(r'^\d+\.\d+\.\d+SNAPSHOT\w+$')

TAG_NOT_ON_HEAD_VERSION_PATTERN = re.compile(r'^\d+\.\d+\.\d+SHA\w+$')
TAG_NOT_ON_HEAD_DIRTY_VERSION_PATTERN = re.compile(r'^\d+\.\d+\.\d+SHA\w+SNAPSHOT\w+$')

here = str(Path(__file__).absolute()).split(os.sep)
bf_path_index = here.index('bigflow')
bf_path_parts = here[:bf_path_index + 1]
BIGFLOW_PATH = os.path.join(os.sep, *bf_path_parts)


class Project:
    def __init__(self):
        self.tmp_dir = Path(tempfile.gettempdir())
        self.project_dir = resolve(self.tmp_dir / f'bigflow_test_version_{str(uuid1()).replace("-", "")}')
        os.mkdir(self.project_dir)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        rmtree(self.project_dir)

    def __del__(self):
        rmtree(self.project_dir, ignore_errors=True)

    def run_cmd(self, cmd):
        result = subprocess.getoutput(f'cd {self.project_dir};{cmd}')
        print(result)
        return result

    def get_version(self):
        get_version_cmd = f"""python -c 'import sys;sys.path.insert(0, "{BIGFLOW_PATH}");from bigflow.version import get_version;print(get_version())'"""
        result = subprocess.getoutput(f'cd {self.project_dir};{get_version_cmd}')
        print(result)
        return result


class GetVersionE2E(TestCase):
    def test_should_version_based_on_git_tags(self):
        with Project() as project:
            # expect
            self.assertTrue(NO_REPOSITORY_VERSION_PATTERN.match(project.get_version()))

            # when
            project.run_cmd('git init')

            # then
            self.assertTrue(NO_COMMITS_VERSION_PATTERN.match(project.get_version()))

            # when
            project.run_cmd("touch file1;git add file1;git commit -m 'file1'")

            # then
            self.assertTrue(NO_TAG_VERSION_PATTERN.match(project.get_version()))

            # when
            project.run_cmd('touch file2')

            # then
            self.assertTrue(NO_TAG_DIRTY_VERSION_PATTERN.match(project.get_version()))

            # when
            project.run_cmd("git add file2;git commit -m 'file2';git tag 0.2.0")

            # then
            self.assertTrue(TAG_ON_HEAD_VERSION_PATTERN.match(project.get_version()))

            # when
            project.run_cmd('touch file3')

            # then
            self.assertTrue(TAG_ON_HEAD_DIRTY_VERSION_PATTERN.match(project.get_version()))

            # when
            project.run_cmd("git add file3;git commit -m 'file3'")

            # then
            self.assertTrue(TAG_NOT_ON_HEAD_VERSION_PATTERN.match(project.get_version()))

            # when
            project.run_cmd('touch file4')

            # then
            self.assertTrue(TAG_NOT_ON_HEAD_DIRTY_VERSION_PATTERN.match(project.get_version()))


class ReleaseTestCase(TestCase):
    @mock.patch('bigflow.version.push_tag')
    @mock.patch('bigflow.version.get_tag')
    def test_should_push_bumped_tag(self, get_tag_mock, push_tag_mock):
        # given
        get_tag_mock.return_value = None

        # when
        release('fake_pem_path')

        # then
        push_tag_mock.assert_called_with(STARTING_VERSION, 'fake_pem_path')

        # given
        get_tag_mock.return_value = '0.2.0'

        # when
        release('fake_pem_path')

        # then
        push_tag_mock.assert_called_with('0.3.0', 'fake_pem_path')


class VersionPatternTestCase(TestCase):
    def test_version_patter(self):
        self.assertTrue(VERSION_PATTERN.match('1.0.0'))
        self.assertTrue(VERSION_PATTERN.match('1.0.1'))
        self.assertTrue(VERSION_PATTERN.match('1.11.1'))
        self.assertTrue(VERSION_PATTERN.match('0.0.1123123'))
        self.assertTrue(VERSION_PATTERN.match('0.0.112dev'))
        self.assertTrue(VERSION_PATTERN.match('0.0.dev'))
        self.assertFalse(VERSION_PATTERN.match('x.0.1123123'))
        self.assertFalse(VERSION_PATTERN.match('x.x.1123123'))
        self.assertFalse(VERSION_PATTERN.match('0.x.1123123'))


class BumpMinorTestCase(TestCase):
    def test_should_bump_minor_(self):
        self.assertEqual(bump_minor('1.0.0'), '1.1.0')
        self.assertEqual(bump_minor('0.1.0'), '0.2.0')
        self.assertEqual(bump_minor('0.1.1'), '0.2.0')
        self.assertEqual(bump_minor('0.0.1'), '0.1.0')
        self.assertEqual(bump_minor('0.1.dev1'), '0.2.0')

    def test_should_raise_value_error_for_invalid_version_schema(self):
        # given
        invalid_version = 'dev.0.1'

        # then
        with self.assertRaises(ValueError):
            # when
            bump_minor(invalid_version)


import os
import json
import shutil
import datetime
import subprocess
from datetime import timedelta

from pathlib import Path
from unittest import TestCase, mock

from bigflow.cli import walk_module_files, SETUP_VALIDATION_MESSAGE
from bigflow.build import now, get_docker_image_id, build_docker_image_tag, auto_configuration, \
    get_docker_repository_from_deployment_config, project_setup, secure_get_version, default_project_setup, \
    clear_image_leftovers, clear_dags_leftovers, clear_package_leftovers, build_image
from example_project.project_setup import DOCKER_REPOSITORY, PROJECT_NAME

TEST_PROJECT_PATH = Path(__file__).parent / 'example_project'
IMAGE_DIR_PATH = TEST_PROJECT_PATH / 'image'
DAGS_DIR_PATH = TEST_PROJECT_PATH / '.dags'
DIST_DIR_PATH = TEST_PROJECT_PATH / 'dist'
EGGS_DIR_PATH = TEST_PROJECT_PATH / f'{PROJECT_NAME}.egg-info'
BUILD_PATH = TEST_PROJECT_PATH / 'build'


def resolve(path: Path):
    return str(path.absolute())


class TestProject:
    def run_build(self, cmd: str):
        output = subprocess.getoutput(f'cd {resolve(TEST_PROJECT_PATH)};{cmd}')
        print(output)
        return output


def mkdir(dir_path: Path):
    if not os.path.isdir(dir_path):
        os.mkdir(dir_path)


def rmdir(dir_path: Path):
    if dir_path.exists() and dir_path.is_dir():
        shutil.rmtree(dir_path)


def create_image_leftovers(test_project_dir_path: Path = TEST_PROJECT_PATH):
    mkdir(test_project_dir_path / 'image')
    (test_project_dir_path / 'image' / 'leftover').touch()


def create_package_leftovers(
        test_project_dir_path: Path = TEST_PROJECT_PATH,
        project_name: str = PROJECT_NAME):
    mkdir(test_project_dir_path / 'build')
    (test_project_dir_path / 'build' / 'leftover').touch()
    mkdir(test_project_dir_path / 'dist')
    (test_project_dir_path / 'dist' / 'leftover').touch()
    mkdir(test_project_dir_path / f'{project_name}.egg-info')
    (test_project_dir_path / f'{project_name}.egg-info' / 'leftover').touch()


def create_dags_leftovers(test_project_dir_path: Path = TEST_PROJECT_PATH):
    mkdir(test_project_dir_path / '.dags')
    (test_project_dir_path / '.dags' / 'leftover').touch()


def dags_leftovers_exist(test_project_dir_path: Path = TEST_PROJECT_PATH):
    return os.path.exists(test_project_dir_path / '.dags' / 'leftover')


def package_leftovers_exist(
        test_project_dir_path: Path = TEST_PROJECT_PATH,
        project_name: str = PROJECT_NAME):
    return os.path.exists(test_project_dir_path / 'build' / 'leftover') and\
           os.path.exists(test_project_dir_path / 'dist' / 'leftover') and\
           os.path.exists(test_project_dir_path / f'{project_name}.egg-info' / 'leftover')


def image_leftovers_exist(test_project_dir_path: Path = TEST_PROJECT_PATH):
    return os.path.exists(test_project_dir_path / 'image' / 'leftover')


def dir_not_empty(dir_path: Path):
    return len(os.listdir(dir_path)) != 0


def file_exists(file_path: Path):
    return os.path.exists(str(file_path.absolute()))


def file_exists_with_json_body(file_path: Path, body: dict):
    if not file_exists(file_path):
        return False
    with open(file_path, 'r') as f:
        return body == json.loads(f.read())


def deployment_config_copied(test_project_dir_path: Path = TEST_PROJECT_PATH):
    return file_exists(test_project_dir_path / 'image' / 'deployment_config.py')


def python_package_built(test_project_dir_path: Path = TEST_PROJECT_PATH):
    return file_exists(test_project_dir_path / 'dist' / 'main_package-0.1.0-py3-none-any.whl')


def test_run(test_project_dir_path: Path = TEST_PROJECT_PATH):
    return file_exists(test_project_dir_path / 'build' / 'junit-reports')


def dags_built(test_project_dir_path: Path, expected_workflow_count: int):
    if not file_exists(test_project_dir_path / '.dags'):
        return False
    return sum(1 for workflow_dir, workflow_name in walk_module_files(test_project_dir_path / '.dags')
                 if 'workflow' in workflow_name) == expected_workflow_count


def docker_image_as_file_built(test_project_dir_path: Path = TEST_PROJECT_PATH):
    return file_exists(test_project_dir_path / 'image' / 'image-0.1.0.tar')


def docker_image_built_in_registry(docker_repository: str, version: str):
    return get_docker_image_id(build_docker_image_tag(docker_repository, version))


def dags_contain(test_project_dir_path: Path, substring: str):
    for module_dir, module_file_name in walk_module_files(test_project_dir_path / '.dags'):
        with open(os.path.join(module_dir, module_file_name), 'r') as f:
            content = f.read()
            if substring not in content:
                return False
    return True


class SetupTestCase(TestCase):
    def setUp(self) -> None:
        self.test_project = TestProject()


class BuildProjectE2E(SetupTestCase):

    def test_should_build_project_artifacts(self):
        # when
        create_package_leftovers()
        create_image_leftovers()
        create_dags_leftovers()
        self.test_project.run_build('python project_setup.py build_project')

        # then
        self.assertTrue(python_package_built())
        self.assertTrue(test_run())
        self.assertTrue(dags_built(TEST_PROJECT_PATH, 2))
        self.assertTrue(docker_image_as_file_built())
        self.assertFalse(docker_image_built_in_registry(DOCKER_REPOSITORY, '0.1.0'))
        self.assertTrue(deployment_config_copied())

        # and
        self.assertFalse(package_leftovers_exist())
        self.assertFalse(image_leftovers_exist())
        self.assertFalse(dags_leftovers_exist())

    def test_should_validate_project_setup(self):
        # expected
        self.assertTrue(SETUP_VALIDATION_MESSAGE in self.test_project.run_build('python project_setup.py build_project --validate-project-setup'))
        self.assertTrue(SETUP_VALIDATION_MESSAGE in self.test_project.run_build('python project_setup.py build_project --build-dags --validate-project-setup'))
        self.assertTrue(SETUP_VALIDATION_MESSAGE in self.test_project.run_build('python project_setup.py build_project --build-image --validate-project-setup'))
        self.assertTrue(SETUP_VALIDATION_MESSAGE in self.test_project.run_build('python project_setup.py build_project --build-package --validate-project-setup'))


class BuildPackageCommandE2E(SetupTestCase):
    def test_should_execute_build_package_command(self):
        # given
        create_package_leftovers()
        clear_image_leftovers(IMAGE_DIR_PATH)
        clear_dags_leftovers(DAGS_DIR_PATH)

        # when
        self.test_project.run_build('python project_setup.py build_project --build-package')

        # then
        self.assertTrue(python_package_built())
        self.assertTrue(test_run())
        self.assertFalse(package_leftovers_exist())


class BuildDagsCommandE2E(SetupTestCase):

    def test_should_execute_build_dags_command(self):
        # given
        create_dags_leftovers()
        clear_image_leftovers(IMAGE_DIR_PATH)
        clear_package_leftovers(DIST_DIR_PATH, EGGS_DIR_PATH, BUILD_PATH)

        # when
        self.test_project.run_build('python project_setup.py build_project --build-dags')

        # then
        self.assertTrue(dags_built(TEST_PROJECT_PATH, 2))
        self.assertFalse(dags_leftovers_exist(TEST_PROJECT_PATH))
        self.assertTrue(dags_contain(TEST_PROJECT_PATH, repr(datetime.datetime.now().replace(hour=0, minute=0, second=0, microsecond=0) - timedelta(hours=24))))

        # when
        self.test_project.run_build("python project_setup.py build_project --build-dags --start-time '2020-01-02 00:00:00'")

        # then
        self.assertTrue(dags_contain(TEST_PROJECT_PATH, 'datetime.datetime(2020, 1, 1, 0, 0)'))

        # when
        self.test_project.run_build('python project_setup.py build_project --build-dags --workflow workflow1')

        # then
        self.assertTrue(self.single_dag_for_workflow_exists('workflow1'))

    def single_dag_for_workflow_exists(self, workflow_id):
        dags_created = dir_not_empty(TEST_PROJECT_PATH / '.dags')
        if not dags_created:
            return False
        dag_file_path = os.path.join(*next(walk_module_files(TEST_PROJECT_PATH / '.dags')))
        with open(dag_file_path, 'r') as f:
            return workflow_id in f.read()


class BuildImageCommandE2E(SetupTestCase):

    def test_should_execute_build_image_command(self):
        # given
        create_image_leftovers()
        clear_dags_leftovers(DAGS_DIR_PATH)
        clear_package_leftovers(DIST_DIR_PATH, EGGS_DIR_PATH, BUILD_PATH)
        self.test_project.run_build('python project_setup.py build_project --build-package')

        # when
        self.test_project.run_build('python project_setup.py build_project --build-image')

        # then
        self.assertFalse(image_leftovers_exist())
        self.assertFalse(docker_image_built_in_registry(DOCKER_REPOSITORY, '0.1.0'))
        self.assertTrue(docker_image_as_file_built())
        self.assertTrue(deployment_config_copied())


class BuildImageTest(TestCase):

    @mock.patch('bigflow.build.build_docker_image_tag')
    @mock.patch('bigflow.build.build_docker_image')
    @mock.patch('bigflow.build.export_docker_image_to_file')
    @mock.patch('bigflow.build.remove_docker_image_from_local_registry')
    def test_should_remove_image_from_local_registry_when_export_to_image_failed(self,
                                                                                 remove_docker_image_from_local_registry,
                                                                                 export_docker_image_to_file,
                                                                                 build_docker_image,
                                                                                 build_docker_image_tag):
        # given
        image_dir_path = Path('image_dir')
        rmdir(image_dir_path)
        export_docker_image_to_file.side_effect = Exception('export failed')
        build_docker_image_tag.return_value = 'tag_value'


        # when:
        with self.assertRaises(Exception):
            build_image('docker_repository', 'v1', Path('project_dir'), image_dir_path, Path('deployment_config'))

        # then
        build_docker_image.assert_called_with(Path('project_dir'), 'tag_value')
        remove_docker_image_from_local_registry.assert_called_with('tag_value')


class AutoConfigurationTestCase(TestCase):
    @mock.patch('bigflow.build.get_version')
    def test_should_produce_default_configuration_for_project_setup(self, get_version_mock):
        # given
        get_version_mock.return_value = '0.1.0'
        project_dir = Path(__file__).parent / 'example_project'

        # expected
        self.assertEqual(auto_configuration('example_project', project_dir), {
            'project_name': 'example_project',
            'docker_repository': 'test_repository',
            'root_package': project_dir / 'example_project',
            'project_dir': project_dir,
            'build_dir': project_dir / 'build',
            'test_package': project_dir / 'test',
            'dags_dir': project_dir / '.dags',
            'dist_dir': project_dir / 'dist',
            'image_dir': project_dir / 'image',
            'eggs_dir': project_dir / 'example_project.egg-info',
            'deployment_config_file': project_dir / 'deployment_config.py',
            'version': '0.1.0',
            'resources_dir': project_dir / 'resources',
            'project_requirements_file': project_dir / 'resources' / 'requirements.txt'
        })

    def test_should_raise_error_when_cant_find_deployment_config(self):
        # then
        with self.assertRaises(ValueError) as e:
            # when
            get_docker_repository_from_deployment_config(Path(__file__).parent / 'example_project' / 'unknown.py')

    @mock.patch('bigflow.build.get_version')
    def test_should_raise_error_when_cant_get_version(self, get_version_mock):
        # given
        get_version_mock.side_effect = self.get_version_error

        # then
        with self.assertRaises(ValueError) as e:
            # when
            secure_get_version()

    def test_should_suite_project_setup(self):
        # given
        project_dir = Path(__file__).parent / 'example_project'

        # expected
        self.assertTrue(project_setup(**auto_configuration('example_project', project_dir)))

    @mock.patch('bigflow.build.get_version')
    @mock.patch('bigflow.build.setup')
    @mock.patch('bigflow.build.build_command')
    def test_should_produce_default_project_setup_using_autoconfiguration(self, build_command_mock, setup_mock, get_version_mock):
        # given
        get_version_mock.return_value = '0.1.0'
        build_command_mock.return_value = 1

        # when
        default_project_setup('example_project',  Path(__file__).parent / 'example_project')

        # then
        setup_mock.assert_called_with(**project_setup(**auto_configuration('example_project',  Path(__file__).parent / 'example_project')))

    def get_version_error(self):
        raise RuntimeError('get_version error')



import logging
import time
import datetime
import functools

import googleapiclient.discovery

logger = logging.getLogger(__name__)


BIGFLOW_JOB_FAILURE_METRIC_TYPE = 'custom.googleapis.com/bigflow'
BIGFLOW_JOB_FAILURE_METRIC = {
        "type": BIGFLOW_JOB_FAILURE_METRIC_TYPE,
        "labels": [
            {
                "key": "job_id",
                "valueType": "STRING",
                "description": "Job UUID"
            }
        ],
        "metricKind": "CUMULATIVE",
        "valueType": "INT64",
        "unit": "items",
        "description": "Job failure counter",
        "displayName": "Job failure count"
    }


class MetricError(RuntimeError):
    pass


def api_list_metrics(client, project_resource, metric_type):
    request = client.projects().metricDescriptors().list(
        name=project_resource,
        filter='metric.type=starts_with("{}")'.format(metric_type))
    return request.execute()


def api_create_timeseries(client, project_resource, data):
    request = client.projects().timeSeries().create(
        name=project_resource,
        body={"timeSeries": [data]})
    return request.execute()


def api_client():
    return googleapiclient.discovery.build('monitoring', 'v3')


def api_create_metric(client, project_resource, metric_descriptor):
    return client.projects().metricDescriptors().create(
        name=project_resource, body=metric_descriptor).execute()


def format_rfc3339(dt):
    return dt.isoformat("T") + "Z"


def get_start_time():
    start_time = datetime.datetime.utcnow() - datetime.timedelta(minutes=5)
    return format_rfc3339(start_time)


def get_now_rfc3339():
    return format_rfc3339(datetime.datetime.utcnow())


def metric_exists(client, project_resource, metric_type):
    response = api_list_metrics(client, project_resource, metric_type)
    try:
        return response['metricDescriptors'] is not None
    except KeyError:
        logger.warning('Metric not found: {}'.format(metric_type))
        return False


def wait_for_metric(client, project_resource, metric_type):
    retries_left = 10
    while not metric_exists(client, project_resource, metric_type) and retries_left:
        logger.info('Waiting for metric: {}'.format(metric_type))
        time.sleep(1)
        retries_left -= 1

    if not retries_left:
        raise MetricError('Waiting for metric {} failed. Project resource {}.'.format(metric_type, project_resource))
    else:
        return True


def create_timeseries_data(metric_type, job_id, project_id, region, environment_name, start, end):
    return {
        "metric": {
            "type": metric_type,
            "labels": {
                "job_id": job_id
            }
        },
        "resource": {
            "type": 'cloud_composer_environment',
            "labels": {
                'project_id': project_id,
                'location': region,
                'environment_name': environment_name
            }
        },
        "points": [
            {
                "interval": {
                    "startTime": start,
                    "endTime": end
                },
                "value": {
                    "int64Value": 1
                }
            }
        ]
    }


def increment_counter(client, monitoring_config, metric_type, job_id):
    start = get_now_rfc3339()
    end = get_now_rfc3339()
    timeseries_data = create_timeseries_data(
        metric_type,
        job_id,
        monitoring_config.project_id,
        monitoring_config.region,
        monitoring_config.environment_name,
        start,
        end)
    api_create_timeseries(client, monitoring_config, timeseries_data)


def increment_job_failure_count(monitoring_config, job_id):
    try:
        client = api_client()
        if not metric_exists(client, monitoring_config.project_resource, BIGFLOW_JOB_FAILURE_METRIC_TYPE):
            api_create_metric(client, monitoring_config.project_resource, BIGFLOW_JOB_FAILURE_METRIC)
            wait_for_metric(client, monitoring_config.project_resource, BIGFLOW_JOB_FAILURE_METRIC_TYPE)
        increment_counter(client, monitoring_config, BIGFLOW_JOB_FAILURE_METRIC_TYPE, job_id)
    except Exception as e:
        raise MetricError('Cannot increment job failure count: ' + str(e)) from e


class MonitoringConfig(object):
    def __init__(self, project_id, region, environment_name):
        self.project_resource = 'projects/{}'.format(project_id)
        self.project_id = project_id
        self.region = region
        self.environment_name = environment_name


def meter_job_run_failures(job, monitoring_config):

    def meter(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except Exception as e:
                logger.exception(e)
                increment_job_failure_count(monitoring_config, job.id)
                raise e
        return wrapper

    # TODO(anjensan): Don't use direct monkey-patching
    if hasattr(job, 'run'):
        job.run = meter(job.run)
    if hasattr(job, 'execute'):
        job.execute = meter(job.execute)

    return job



import re
from typing import Optional
from uuid import uuid1
import subprocess

from better_setuptools_git_version import get_tag
from better_setuptools_git_version import get_version as base_get_version

VERSION_PATTERN = re.compile(r'^(\d+\.)?(\d+\.)?(\w+)$')

__all__ = [
    'get_version',
    'release'
]

STARTING_VERSION = '0.1.0'


def get_version() -> str:
    """
    case 1: no .git / no commits / error
        0.1.0 + uuid
    case 2: no tags:
        0.1.0 + {uuid if dirty}
    case 2: tag on head
        tag + {uuid if dirty}
    case 4: tag not on head
        last tag + sha + {uuid if dirty}
    """
    result = base_get_version(
        template="{tag}SHA{sha}",
        starting_version=STARTING_VERSION).replace('+dirty', f'SNAPSHOT{short_uuid()}')
    if not VERSION_PATTERN.match(result):
        return f'{STARTING_VERSION}SNAPSHOT{short_uuid()}'
    if result == STARTING_VERSION:
        return f"{STARTING_VERSION}SNAPSHOT{short_uuid()}"
    return result


def short_uuid():
    return str(uuid1()).replace("-", "")[:8]


def release(identity_file: Optional[str] = None) -> None:
    latest_tag = get_tag()
    if latest_tag:
        tag = bump_minor(latest_tag)
    else:
        tag = '0.1.0'
    push_tag(tag, identity_file)


def push_tag(tag, identity_file: Optional[str] = None) -> None:
    print(f'Setting and pushing tag: {tag}')
    print(subprocess.getoutput(f'git tag {tag}'))
    if identity_file is not None:
        print(f'Pushing using the specified identity_file: {identity_file}')
        print(subprocess.getoutput(
            f"GIT_SSH_COMMAND='ssh -i {identity_file} -o IdentitiesOnly=yes' git push origin {tag}"))
    else:
        print(subprocess.getoutput('git push origin --tags'))


def bump_minor(version: str) -> str:
    if not VERSION_PATTERN.match(version):
        raise ValueError('Expected version pattern is <major: int>.<minor: int>.<patch: int>.')
    major, minor, patch = version.split('.')
    minor = str(int(minor) + 1)
    return f'{major}.{minor}.0'



try:
    from .interactive import interactive_component as component
    from .interactive import add_label, sensor, INLINE_COMPONENT_DATASET_ALIAS
    from .dataset_configuration import DatasetConfig
    from .job import Job
    from .interface import Dataset
    __all__ = [
        'component',
        'add_label',
        'sensor',
        'INLINE_COMPONENT_DATASET_ALIAS',
        'DatasetConfig',
        'Job'
    ]
except ImportError:
    __all__ = []




from abc import ABCMeta, abstractmethod
from .job import DEFAULT_RETRY_COUNT
from .job import DEFAULT_RETRY_PAUSE_SEC

DEFAULT_RUNTIME = '1970-01-01'


class BigQueryOperation(object, metaclass=ABCMeta):

    @abstractmethod
    def to_job(self, id: str, retry_count: int = DEFAULT_RETRY_COUNT, retry_pause_sec: int = DEFAULT_RETRY_PAUSE_SEC):
        pass

    def run(self, runtime=DEFAULT_RUNTIME):
        pass


class Dataset(object, metaclass=ABCMeta):

    @abstractmethod
    def write_truncate(self, table_name: str, sql: str, partitioned: bool = True) -> BigQueryOperation:
        pass

    @abstractmethod
    def write_append(self, table_name: str, sql: str, partitioned: bool = True) -> BigQueryOperation:
        pass

    @abstractmethod
    def write_tmp(self, table_name: str, sql: str) -> BigQueryOperation:
        pass

    @abstractmethod
    def collect(self, sql: str) -> BigQueryOperation:
        pass

    @abstractmethod
    def collect_list(self, sql: str) -> BigQueryOperation:
        pass

    @abstractmethod
    def create_table(self, create_query: str) -> BigQueryOperation:
        pass



import base64
import functools
import datetime
import time
import inspect
import io
import os
import pathlib
import pickle
import random
import re
import string
import subprocess
import textwrap
import time
import typing
import contextlib
import logging

from google.cloud import dataproc_v1, storage

import bigflow
import bigflow.configuration
import bigflow.resources
import bigflow.commons


logger = logging.getLogger(__name__)

__all__ = [
    'PySparkJob',
]

DEFAULT_REQUIREMENTS = [
    f"bigflow[dataproc,log,bigquery]=={bigflow.__version__}",
]


class PySparkJob:

    def __init__(
        self,
        id,
        driver_callable: typing.Callable,
        bucket_id: str,
        gcp_project_id: str,
        gcp_region: str,
        driver_arguments: typing.Optional[dict] = None,
        pip_packages: typing.Union[typing.Iterable[str], pathlib.Path] = (),
        jar_file_uris: typing.Iterable[str] = ('gs://spark-lib/bigquery/spark-bigquery-latest_2.12.jar',),
        worker_num_instances: int = 2,
        worker_machine_type: str = 'n1-standard-1',
        env: typing.Optional[str] = None,
        setup_file: typing.Optional[str] = None,
    ):
        self.id = id

        self.driver_filename = 'driver.py'
        self.driver_callable = driver_callable
        self.driver_arguments = driver_arguments or {}

        self.bucket_id = bucket_id
        self.gcp_project_id = gcp_project_id
        self.gcp_region = gcp_region

        self.env = env or bigflow.configuration.current_env() or 'none'

        self.jar_file_uris = jar_file_uris
        self.worker_machine_type = worker_machine_type
        self.worker_num_instances = worker_num_instances

        if isinstance(pip_packages, pathlib.Path):
            self.pip_packages = bigflow.resources.read_requirements(pip_packages)
        elif pip_packages:
            self.pip_packages = pip_packages
        else:
            self.pip_packages = DEFAULT_REQUIREMENTS

        if setup_file:
            self.setup_file = setup_file
        else:
            self.setup_file = None
            self._project = _capture_caller_topmodule()
            self._any_file_inside_project = _capture_caller_path(1)

    def _ensure_has_setup_file(self):
        if self.setup_file:
            return
        logger.debug("Find or create setup file for main package")
        self.setup_file = bigflow.resources.find_or_create_setup_for_main_project_package(
            self._project,
            self._any_file_inside_project)

    def _generate_internal_jobid(self, runtime):
        job_random_suffix = ''.join(random.choices(string.ascii_lowercase + string.digits, k=10))
        job_timestamp = datetime.datetime.now().strftime("%Y-%m-%d-%H%M%S")  # FIXME: use 'runtime' ?
        jobid = f"{self.id.lower()}-{self.env or 'none'}-{job_timestamp}-{job_random_suffix}"
        logger.info("Internal job is %r", jobid)
        return jobid

    def _prepare_driver_callable(self, runtime):
        kwargs = dict(self.driver_arguments)
        kwargs['runtime'] = runtime
        
        envs = {}
        if self.env:
            logger.debug("Will set env variable 'env' on pyspark driver to %r", self.env)
            envs['env'] = self.env

        return _PySparkDriverWrapper(self.driver_callable, (), kwargs, envs)

    @contextlib.contextmanager
    def _with_temp_cluster(self, cluster_name):
        client_options = {'api_endpoint': f"{self.gcp_region}-dataproc.googleapis.com:443"}
        dataproc_cluster_client = dataproc_v1.ClusterControllerClient(client_options=client_options)
        try:
            logger.debug("Create temp cluster %r", cluster_name)
            cluster_name = _create_cluster(
                dataproc_cluster_client=dataproc_cluster_client,
                project_id=self.gcp_project_id,
                region=self.gcp_region,
                cluster_name=cluster_name,
                requirements=self.pip_packages,
                worker_machine_type=self.worker_machine_type,
                worker_num_instances=self.worker_num_instances,
            )
            yield cluster_name
        finally:
            logger.debug("Delete temp cluster %r", cluster_name)
            _delete_cluster(dataproc_cluster_client, self.gcp_project_id, self.gcp_region, cluster_name)

    def _prepare_pyspark_properties(self):
        ps = {
            'spark.app.name': self.id,
        }
        if self.env:
            ps['spark.workerEnv.bf_env'] = self.env            
            ps['spark.executorEnv.bf_env'] = self.env
        logger.debug("Pyspark properties are %r", ps)
        return ps

    def run(self, runtime):
        logger.info("Run job %r", self.id)

        self._ensure_has_setup_file()
        job_internal_id = self._generate_internal_jobid(runtime)

        client_options = {'api_endpoint': f"{self.gcp_region}-dataproc.googleapis.com:443"}
        storage_client = storage.Client(project=self.gcp_project_id)
        dataproc_job_client = dataproc_v1.JobControllerClient(client_options=client_options)

        driver = self._prepare_driver_callable(runtime)

        logger.info("Prapare and upload python package...")
        bucket = storage_client.get_bucket(self.bucket_id)
        egg_local_path = _build_project_egg(self.setup_file)
        egg_path = _upload_egg(egg_local_path, bucket, job_internal_id)

        driver_path = f"{job_internal_id}/{self.driver_filename}"
        generate_and_upload_driver(driver, bucket, driver_path)

        with self._with_temp_cluster(job_internal_id) as cluster_name:
            job = _submit_single_pyspark_job(
                dataproc_job_client=dataproc_job_client,
                project_id=self.gcp_project_id,
                region=self.gcp_region,
                cluster_name=cluster_name,
                bucket_id=self.bucket_id,
                jar_file_uris=self.jar_file_uris,
                driver_path=driver_path,
                egg_path=egg_path,
                properties=self._prepare_pyspark_properties(),
            )
            try:
                _wait_for_job_to_finish(dataproc_job_client, self.gcp_project_id, self.gcp_region, job)
            finally:
                _print_job_output_log(storage_client, dataproc_job_client, self.gcp_project_id, self.gcp_region, job)

        logger.info("Job %r was finished", self.id)


class _PySparkDriverWrapper:

    def __init__(self, callable, args, kwargs, envs):
        self.callable = callable
        self.args = args
        self.kwargs = kwargs
        self.envs = envs

    def __call__(self):
        # TODO(anjensan): Configure logging (depends on #90)
        os.environ.update(self.envs)
        self.callable(*self.args, **self.kwargs)


def generate_driver_script(callable):
    pickled = pickle.dumps(callable)
    return textwrap.dedent(f"""

        # Generated by `bigflow.dataproc.generate_and_upload_driver`
        import pickle, base64
        
        # unpickle'n'call {callable!r}
        data = {base64.b85encode(pickled)!r}

        call = pickle.loads(base64.b85decode(data))
        call()
    
    """)


def generate_and_upload_driver(callable, bucket, driver_path):
    logger.debug("Upload generate driver to %s", driver_path)
    driver_blob = bucket.blob(driver_path)
    script = generate_driver_script(callable)
    driver_blob.upload_from_string(content_type='application/octet-stream', data=script)


def _upload_egg(egg_local_path, bucket, job_internal_id):
    logger.debug("Upload egg to %s", egg_local_path)
    egg_path = '{}/{}'.format(job_internal_id, pathlib.Path(egg_local_path).name)
    egg_blob = bucket.blob(egg_path)
    egg_blob.upload_from_filename(filename=egg_local_path, content_type='application/octet-stream')
    return egg_path


def _wait_for_job_to_finish(dataproc_cluster_client, project_id, region, job_id):
    logger.info("Waiting for job to finish...")
    while True:
        job = dataproc_cluster_client.get_job(project_id=project_id, region=region, job_id=job_id)
        state = job.status.state
        if state == dataproc_v1.JobStatus.State.DONE:
            logger.info("Job is DONE")
            return job
        elif state == dataproc_v1.JobStatus.State.ERROR:
            logger.error("Job was failed with ERROR")
            raise Exception(job.status.details)
        elif state == dataproc_v1.JobStatus.State.CANCELLED:
            logger.warn("Job was CANCELLED")            
            raise Exception("Job was CANCELLED")
        time.sleep(5)


def _print_job_output_log(storage_client, dataproc_cluster_client, project_id, region, job_id):

    job = dataproc_cluster_client.get_job(project_id=project_id, region=region, job_id=job_id)
    log_buffer = io.BytesIO()
    # FIXME(anjensan): Download all logs
    storage_client.download_blob_to_file(job.driver_output_resource_uri + ".000000000", log_buffer)
    logger.info("JOB OUTPUT:\n=====\n%s\n=====", log_buffer.getvalue().decode())


def _create_cluster(dataproc_cluster_client, project_id, region, cluster_name, requirements, worker_num_instances, worker_machine_type):
    packages = " ".join(filter(None, requirements))
    cluster_data = {
        # "project_id": project_id,
        "cluster_name": cluster_name,
        "config": {
            "master_config": {
                "num_instances": 1,
                "machine_type_uri": "n1-standard-4",
                "disk_config": {
                    "boot_disk_size_gb": 100
                }},
            "worker_config": {
                "num_instances": worker_num_instances,
                "machine_type_uri": worker_machine_type,
                "disk_config": {
                    "boot_disk_size_gb": 100
                }},
            "software_config": {"image_version": "1.5"},
            "initialization_actions": [
                {"executable_file": "gs://goog-dataproc-initialization-actions-{}/python/pip-install.sh".format(region)}
            ],
            "gce_cluster_config": {
                "metadata": [('PIP_PACKAGES', packages)]
            }
        },
    }
    logger.debug("Cluster spec %r", cluster_data)

    start_at = time.time()
    logger.info("Create cluster %r", cluster_name)
    cluster_future = dataproc_cluster_client.create_cluster(
        project_id=project_id,
        region=region,
        cluster=cluster_data,
    )
    
    logger.info("Waiting for cluster creation...")
    cluster_future.result()

    passed = time.time() - start_at
    logger.info("Cluster created in %s seconds." % passed)

    return cluster_name


def _submit_single_pyspark_job(
        dataproc_job_client, 
        project_id,
        region,
        cluster_name,
        bucket_id,
        driver_path,
        egg_path,
        jar_file_uris,
        properties,
    ):
    jobspec = {
        "placement": {
            "cluster_name": cluster_name,
        },
        "pyspark_job": {
            'main_python_file_uri': f"gs://{bucket_id}/{driver_path}",
            'jar_file_uris': jar_file_uris,
            'python_file_uris': [f"gs://{bucket_id}/{egg_path}"],
            'properties': properties,
        },
    }
    logger.debug("Pyspark job %r", jobspec)
    result = dataproc_job_client.submit_job(project_id=project_id, region=region, job=jobspec)

    job_id = result.reference.job_id
    print("Job {} submitted.".format(job_id))
    print(f"https://console.cloud.google.com/dataproc/jobs/{job_id}?project={project_id}&region={region}")

    return job_id


def _delete_cluster(dataproc_cluster_client, project_id, region, cluster):
    print("Deleting cluster...")
    dataproc_cluster_client.delete_cluster(
        project_id=project_id,
        region=region,
        cluster_name=cluster,
    )
    print("Cluster was deleted.")


def _capture_caller_path(deep=1):
    return os.path.abspath(inspect.stack()[deep + 1].filename)


def _capture_caller_topmodule(deep=1):
    f = inspect.stack()[deep + 1].frame
    module = f.f_globals['__name__']
    return module.split(".", 2)[0]


def _build_project_egg(setup_file):
    cwd = pathlib.Path(setup_file).parent.resolve()
    output = bigflow.commons.run_process(
        ['python', setup_file, 'bdist_egg'], 
        cwd=str(cwd),
    )
    egg_filename = re.search('creating \'([^\']*)\'', output).group(1)
    logger.debug("Egg filename is %r", egg_filename)
    return str(cwd / egg_filename)



import logging
import contextlib
import io
import sys

from unittest import TestCase, mock

from google.cloud import logging_v2

import bigflow.log


class LoggerTestCase(TestCase):

    def configure_mocked_logging(self, project_id, log_name, workflow_id=None):
        self.logging_client = mock.MagicMock(return_value=mock.create_autospec(logging_v2.LoggingServiceV2Client))
        with mock.patch.object(bigflow.log.GCPLoggerHandler, 'create_logging_client', return_value=self.logging_client):
            bigflow.log._LOGGING_CONFIGURED = False
            bigflow.log.init_logging(
                workflow_id=workflow_id,
                config={
                    'gcp_project_id': project_id,
                    'log_name': log_name,
                },
            )

    def setUp(self):
        self.configure_mocked_logging('project-id', 'logger_name', 'workflow-id')
        self.test_logger = logging.getLogger('any.random.logger.name')
        self.root_logger = logging.getLogger('')

    def _clear_all_root_loggers(self):
        for h in logging.getLogger().handlers[:]:
            logging.getLogger().removeHandler(h)
            h.close()
 
    def tearDown(self):
        self._clear_all_root_loggers()
        bigflow.log._LOGGING_CONFIGURED = False

    def test_should_create_correct_logging_link(self):
        # when
        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            # stdout handler is created only when no other handlers are registered
            self._clear_all_root_loggers()
            self.configure_mocked_logging('project-id', 'another_log_name', 'workflow_id')

        # then
        out = f.getvalue()
        self.assertIn("LOGS LINK", out)
        self.assertIn("https://console.cloud.google.com/logs/query;query=", out)
        self.assertIn("labels.workflow_id%3D%22workflow_id%22", out)

    def test_should_log_to_correct_streams(self):
        # given
        err = io.StringIO()
        out = io.StringIO()

        # when
        with contextlib.redirect_stdout(out):
            self._clear_all_root_loggers()
            self.configure_mocked_logging('project-id', 'another_log_name', 'workflow_id')
            logger = logging.getLogger()
            logger.info("info")
            logger.warning("watch out")
            logger.error("oh no")
            logger.critical("nooo...")

        # then
        out = out.getvalue()
        self.assertIn("info", out)
        self.assertIn("watch out", out)
        self.assertNotIn("oh no", out)
        self.assertNotIn("nooo...", out)

        # when
        with contextlib.redirect_stderr(err):
            self._clear_all_root_loggers()
            self.configure_mocked_logging('project-id', 'another_log_name', 'workflow_id')
            logger = logging.getLogger()
            logger.info("info")
            logger.warning("watch out")
            logger.error("oh no")
            logger.critical("nooo...")

        # then
        err_out = err.getvalue()
        self.assertNotIn("info", err_out)
        self.assertNotIn("watch out", err_out)
        self.assertIn("oh no", err_out)
        self.assertIn("nooo...", err_out)

    def _assert_single_log_event(self, message_re, severity=None):
        self.assertEqual(1, self.logging_client.write_log_entries.call_count)
        calls = self.logging_client.write_log_entries.call_args_list[0][0]
        le = calls[0][0]

        self.assertEqual(le.resource.type, "global")
        self.assertEqual(le.resource.labels, {"project_id": "project-id"})
        self.assertEqual(le.log_name, "projects/project-id/logs/logger_name")
        
        if severity:
            self.assertEqual(le.severity, severity)
        if message_re:
            self.assertRegexpMatches(le.json_payload['message'], message_re)

        return le

    def test_should_log_unhandled_exception(self):

        # when
        try:
            raise ValueError("oh no... i'm dying")
        except Exception:
            sys.excepthook(*sys.exc_info()) # simulate uncaught exception

        # then
        le = self._assert_single_log_event(
            message_re="Uncaught exception: oh no... i\'m dying",
            severity=500,
        )
        self.assertIn("ValueError", le.json_payload['exc_text'])

    def test_should_handle_warning(self):
        # when
        self.test_logger.warning("warning message")

        # then
        self._assert_single_log_event(
            message_re="warning message",
            severity=400,
        )

    def test_should_handle_info(self):
        # when
        self.test_logger.info("info message")

        # then
        self._assert_single_log_event(
            message_re="info message",
            severity=200,
        )

    def test_should_handle_error(self):
        # when
        self.test_logger.error("error message")

        # then
        self._assert_single_log_event(
            message_re="error message",
            severity=500,
        )

    def test_should_install_gcp_handler_when_logging_already_exists(self):

        # given
        self._clear_all_root_loggers()
        logging.basicConfig(level=logging.INFO)

        # when
        self.configure_mocked_logging('project-id', 'logger_name', 'workflow_id')
        self.test_logger.info("message")

        # then
        self._assert_single_log_event(
            message_re="message",
            severity=200,
        )



from unittest import TestCase
from bigflow.bigquery.dataset_manager import handle_key_error
from bigflow.bigquery.dataset_manager import AliasNotFoundError


class HandleKeyErrorTestCase(TestCase):
    def test_should_catch_key_error_and_raise_variable_not_found_error(self):
        # expect
        with self.assertRaises(AliasNotFoundError):
            self.raise_key_error()

    @handle_key_error
    def raise_key_error(self):
        '{missing_key}'.format(meh='bla')


import subprocess
import sys
from pathlib import Path
import logging
from datetime import datetime


logger = logging.getLogger(__name__)


def now(template: str = "%Y-%m-%d %H:00:00"):
    return datetime.now().strftime(template)


def run_process(cmd, **kwargs):
    if isinstance(cmd, str):
        cmd = cmd.split(' ')
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, **kwargs)
    result_output = ''
    for c in iter(lambda: process.stdout.read(1), b''):
        l = c.decode('utf-8')
        sys.stdout.write(c.decode('utf-8'))
        result_output += l
    return result_output


def resolve(path: Path):
    return str(path.absolute())


def not_none_or_error(arg_value, arg_name):
    if arg_value is None:
        raise ValueError("{} can't be None".format(arg_name))


class ExtrasRequiredError(ImportError):
    pass


def merge_dicts(dict1, dict2):
    return {**dict1, **dict2}


def decode_version_number_from_file_name(file_path: Path):
    if file_path.suffix != '.tar':
        raise ValueError(f'*.tar file expected in {file_path.as_posix()}, got {file_path.suffix}')
    if not file_path.is_file():
        raise ValueError(f'File not found: {file_path.as_posix()}')

    split = file_path.stem.split('-', maxsplit=1)
    if not len(split) == 2:
        raise ValueError(f'Invalid file name pattern: {file_path.as_posix()}, expected: *-{{version}}.tar, for example: image-0.1.0.tar')
    return split[1]


def get_docker_image_id(tag):
    images = subprocess.getoutput(f"docker images -q {tag}")
    return images.split('\n')[0]


def build_docker_image_tag(docker_repository: str, package_version: str):
    return docker_repository + ':' + package_version


def remove_docker_image_from_local_registry(tag):
    print('Removing the image from the local registry')
    run_process(f"docker rmi {get_docker_image_id(tag)}")



import shutil
import subprocess
import unittest
from pathlib import Path
from unittest import TestCase, mock

from bigflow.cli import cli


class ProjectScaffoldE2ETestCase(TestCase):
    def tearDown(self):
        clear_project_leftovers(Path('my_project_project'))


class ProjectScaffoldE2E(ProjectScaffoldE2ETestCase):
    @mock.patch('bigflow.cli.gcloud_project_list')
    @mock.patch('bigflow.cli.gcp_bucket_input')
    @mock.patch('bigflow.cli.gcp_project_input')
    @mock.patch('bigflow.cli.project_type_input')
    @mock.patch('bigflow.cli.project_name_input')
    def test_should_create_basic_project(self, project_name_input_mock, project_type_input_mock, gcp_project_input_mock,
                                         gcp_bucket_input_mock, gcloud_project_list_mock):
        # given
        gcloud_project_list_mock.return_value = ''
        project_name_input_mock.return_value = 'my_project'
        project_type_input_mock.return_value = 'b'
        gcp_project_input_mock.return_value = 'my_gcp_project'
        gcp_bucket_input_mock.return_value = 'my_gcp_bucket'

        # when
        cli(['start-project'])

        # then
        self.scaffolded_basic_project_should_have_one_environment()

        # and
        self.scaffolded_project_tests_should_work()

    @unittest.skip('advanced mode not available yet')
    @mock.patch('bigflow.cli.project_number_input')
    @mock.patch('bigflow.cli.environment_name_input')
    @mock.patch('bigflow.cli.gcp_bucket_input')
    @mock.patch('bigflow.cli.gcp_project_input')
    @mock.patch('bigflow.cli.project_type_input')
    @mock.patch('bigflow.cli.project_name_input')
    def test_should_create_advanced_project(self, project_name_input_mock, project_type_input_mock,
                                            gcp_project_input_mock, gcp_bucket_input_mock, environment_name_input,
                                            project_number_input_mock):
        # given
        project_name_input_mock.return_value = 'my_project'
        project_type_input_mock.return_value = 'a'
        gcp_project_input_mock.side_effect = ['dev_project', 'test_project', 'prod_project']
        gcp_bucket_input_mock.side_effect = ['dev_bucket', 'test_bucket', 'prod_bucket']
        environment_name_input.side_effect = ['DEV', 'TEST', 'PROD']
        project_number_input_mock.return_value = '3'

        # when
        cli(['start-project'])

        # then
        self.scaffolded_advanced_project_should_have_three_environments()

        # and
        self.scaffolded_project_tests_should_work()

    def scaffolded_project_tests_should_work(self):
        output = subprocess.getoutput("python -m unittest discover -s my_project_project -p '*.py'")
        self.assertEqual(output[-2:], 'OK')

    def scaffolded_basic_project_should_have_one_environment(self):
        self.check_file_content(Path('my_project_project') / 'deployment_config.py', '''from bigflow.configuration import Config

deployment_config = Config(
    name='dev',
    properties={
       'docker_repository': 'test_repository',
       'gcp_project_id': 'my_gcp_project',
       'dags_bucket': 'my_gcp_bucket'})
''')
        self.check_file_content(Path('my_project_project') / 'my_project' / 'wordcount' / 'config.py', '''from bigflow.configuration import Config

workflow_config = Config(name='dev',
    properties={
        'project_name': 'my_project',
        'gcp_project_id': 'my_gcp_project',
        'staging_location': 'my_project/beam_runner/staging',
        'temp_location': 'my_project/beam_runner/temp',
        'region': 'europe-west1',
        'machine_type': 'n1-standard-1'}).resolve()''')

    def scaffolded_advanced_project_should_have_three_environments(self):
        self.check_file_content(Path('my_project_project') / 'deployment_config.py', '''from bigflow.configuration import Config

deployment_config = Config(name='dev',
                           properties={
                               'docker_repository': 'test_repository',
                               'gcp_project_id': 'dev_project',
                               'dags_bucket': 'dev_bucket'
                           }).add_configuration(name='TEST',
                           properties={
                               'gcp_project_id': 'test_project',
                               'dags_bucket': 'test_bucket'
                           }).add_configuration(name='PROD',
                           properties={
                               'gcp_project_id': 'prod_project',
                               'dags_bucket': 'prod_bucket'
                           })
''')
        self.check_file_content(Path('my_project_project') / 'workflows' / 'internationalports' / 'workflow.py', """from bigflow import Workflow
from bigflow.bigquery import DatasetConfig

dataset_config = DatasetConfig(
    env='dev',
    project_id='dev_project',
    dataset_name='internationalports',
    internal_tables=['ports', 'polish_ports'],
    external_tables={})

dataset = dataset_config.create_dataset_manager()

create_polish_ports_table = dataset.create_table('''
    CREATE TABLE IF NOT EXISTS polish_ports (
      port_name STRING,
      port_latitude FLOAT64,
      port_longitude FLOAT64)
''')

create_ports_table = dataset.create_table('''
    CREATE TABLE IF NOT EXISTS ports (
      port_name STRING,
      port_latitude FLOAT64,
      port_longitude FLOAT64,
      country STRING,
      index_number STRING)
''')

select_polish_ports = dataset.write_truncate('ports', '''
        SELECT port_name, port_latitude, port_longitude
        FROM `{more_ports}`
        WHERE country = 'POL'
        ''', partitioned=False)

populate_ports_table = dataset.collect('''
        INSERT INTO `{more_ports}` (port_name, port_latitude, port_longitude, country, index_number)
        VALUES 
        ('GDYNIA', 54.533333, 18.55, 'POL', '28740'),
        ('GDANSK', 54.35, 18.666667, 'POL', '28710'),
        ('MURMANSK', 68.983333, 33.05, 'RUS', '62950'),
        ('SANKT-PETERBURG', 59.933333, 30.3, 'RUS', '28370');
        ''')


internationalports_workflow = Workflow(
        workflow_id='internationalports',
        definition=[
                create_ports_table.to_job(id='create_ports_table'),
                create_polish_ports_table.to_job(id='create_polish_ports_table'),
                populate_ports_table.to_job(id='populate_ports_table'),
                select_polish_ports.to_job(id='select_polish_ports'),
        ],
        schedule_interval='@once')""")

    def check_file_content(self, path, template):
        with open(path.resolve(), 'r') as f:
            self.assertEqual(f.read(), template)


def clear_project_leftovers(image_dir: Path):
    shutil.rmtree(image_dir, ignore_errors=True)



from unittest import TestCase, mock
from datetime import datetime

import bigflow
from bigflow import monitoring
from freezegun import freeze_time


TEST_DATETIME = datetime(year=2019, month=1, day=2)
TEST_DATETIME_RFC3339 = '2019-01-02T00:00:00Z'


class FormatRFC3339TestCase(TestCase):

    def test_should_format_datetime(self):
        # expect
        self.assertEqual(
            monitoring.format_rfc3339(TEST_DATETIME),
            TEST_DATETIME_RFC3339)


class GetStartTimeTestCase(TestCase):

    @freeze_time(TEST_DATETIME)
    def test_should_return_utc_now_minus_5_minutes(self):
        # expect
        self.assertEqual(
            monitoring.get_start_time(),
            '2019-01-01T23:55:00Z')


class GetNowRFC3339(TestCase):

    @freeze_time(TEST_DATETIME)
    def test_should_return_now_as_rfc3339(self):
        # expect
        self.assertEqual(
            monitoring.get_now_rfc3339(),
            TEST_DATETIME_RFC3339)


class MetricExistsTestCase(TestCase):

    @mock.patch('bigflow.monitoring.api_list_metrics')
    def test_should_check_if_metric_exists(self, api_list_metrics_mock: mock.Mock):
        # given
        api_list_metrics_mock.return_value = {'metricDescriptors': 'something'}

        # expect
        self.assertTrue(monitoring.metric_exists('client', 'project_resource', 'metric_type'))
        api_list_metrics_mock.assert_has_calls([mock.call('client', 'project_resource', 'metric_type')])

        # given
        api_list_metrics_mock.return_value = {}

        # expect
        self.assertFalse(monitoring.metric_exists('client', 'project_resource', 'metric_type'))


class WaitForMetricTestCase(TestCase):

    @mock.patch('bigflow.monitoring.metric_exists')
    def test_should_wait_until_metric_exists(self, metric_exists_mock):
        # given
        metric_exists_mock.return_value = True

        # expect
        self.assertTrue(monitoring.wait_for_metric('client', 'project_resource', 'metric_type'))

    @mock.patch('bigflow.monitoring.metric_exists')
    def test_should_raise_exception_when_no_retries_left(self, metric_exists_mock):
        # given
        metric_exists_mock.return_value = False

        # expect
        with self.assertRaises(monitoring.MetricError):
            monitoring.wait_for_metric('client', 'project_resource', 'metric_type')


class CreateTimeseriesData(TestCase):
    def test_should_return_timeseries_json(self):
        # expect
        self.assertEqual(
            monitoring.create_timeseries_data('metric type', 'job id', 'project id', 'region', 'env name', 'start',
                                              'end'), {
                "metric": {
                    "type": 'metric type',
                    "labels": {
                        "job_id": 'job id'
                    }
                },
                "resource": {
                    "type": 'cloud_composer_environment',
                    "labels": {
                        'project_id': 'project id',
                        'location': 'region',
                        'environment_name': 'env name'
                    }
                },
                "points": [
                    {
                        "interval": {
                            "startTime": 'start',
                            "endTime": 'end'
                        },
                        "value": {
                            "int64Value": 1
                        }
                    }
                ]
            })


class IncrementCounterTestCase(TestCase):

    @freeze_time(TEST_DATETIME)
    @mock.patch('bigflow.monitoring.api_create_timeseries')
    def test_should_call_api_create_timeseries(self, api_create_timeseries_mock: mock.Mock):
        # given
        monitoring_config = monitoring.MonitoringConfig(
            'test project',
            'eu-west',
            'env')

        # when
        monitoring.increment_counter('client', monitoring_config, 'metric type', 'job id')

        # then
        api_create_timeseries_mock.assert_called_once_with(
            'client',
            monitoring_config,
            monitoring.create_timeseries_data(
                'metric type',
                'job id',
                'test project',
                'eu-west',
                'env',
                TEST_DATETIME_RFC3339,
                TEST_DATETIME_RFC3339
            ))


class IncrementJobFailureCount(TestCase):

    @mock.patch('bigflow.monitoring.api_client')
    @mock.patch('bigflow.monitoring.increment_counter')
    @mock.patch('bigflow.monitoring.wait_for_metric')
    @mock.patch('bigflow.monitoring.api_create_metric')
    @mock.patch('bigflow.monitoring.metric_exists')
    def test_should_create_metric_if_not_found(self,
                                               metric_exists_mock,
                                               api_create_metric_mock,
                                               wait_for_metric_mock,
                                               increment_counter_mock,
                                               api_client_mock):
        # given
        monitoring_config = monitoring.MonitoringConfig(
            'test project',
            'eu-west',
            'env')
        api_client_mock.return_value = 'client'
        metric_exists_mock.return_value = False

        # when
        monitoring.increment_job_failure_count(monitoring_config, 'job id')

        # then
        api_create_metric_mock.assert_called_once_with(
            'client',
            monitoring_config.project_resource,
            monitoring.BIGFLOW_JOB_FAILURE_METRIC)

    @mock.patch('bigflow.monitoring.api_client')
    @mock.patch('bigflow.monitoring.increment_counter')
    @mock.patch('bigflow.monitoring.wait_for_metric')
    @mock.patch('bigflow.monitoring.api_create_metric')
    @mock.patch('bigflow.monitoring.metric_exists')
    def test_should_increment_counter(self,
                                      metric_exists_mock,
                                      api_create_metric_mock,
                                      wait_for_metric_mock,
                                      increment_counter_mock,
                                      api_client_mock):
        # given
        monitoring_config = monitoring.MonitoringConfig(
            'test project',
            'eu-west',
            'env')
        api_client_mock.return_value = 'client'
        metric_exists_mock.return_value = True

        # when
        monitoring.increment_job_failure_count(monitoring_config, 'job id')

        # then
        increment_counter_mock.assert_called_once_with(
            'client',
            monitoring_config,
            monitoring.BIGFLOW_JOB_FAILURE_METRIC_TYPE,
            'job id')


class FailingJob(object):
    def __init__(self, id):
        self.id = id

    def execute(self, context):
        raise Exception('Panic!')


class MeterJobRunFailuresTestCase(TestCase):

    @mock.patch('bigflow.monitoring.increment_job_failure_count')
    def test_should_increment_failure_count_when_job_fail(self, increment_job_failure_count_mock):
        # given
        monitoring_config = monitoring.MonitoringConfig(
            'test project',
            'eu-west',
            'env')

        metered_job = monitoring.meter_job_run_failures(FailingJob('job1'), monitoring_config)

        # when
        with self.assertRaises(Exception):
            metered_job.execute(bigflow.JobContext.make(runtime='2019-01-01'))

        # then
        increment_job_failure_count_mock.assert_called_once_with(monitoring_config, 'job1')


from unittest import TestCase, mock
from unittest.mock import patch

from google.cloud import bigquery
from bigflow.bigquery.dataset_manager import DatasetManager, TemplatedDatasetManager
from bigflow.scaffold.scaffold_templates import bq_workflow_template

exec(bq_workflow_template % {'project_id': 'gcp-project'})


class ScaffoldBigQueryExampleTestCase(TestCase):
    @mock.patch('bigflow.bigquery.dataset_manager.create_bigquery_client')
    @patch.object(DatasetManager, 'write')
    @patch.object(DatasetManager, 'create_table')
    @patch.object(DatasetManager, 'collect')
    @patch.object(TemplatedDatasetManager, 'create_full_table_id',
                  side_effect=lambda table: 'gcp-project.bigflow_test' + '.' + table)
    @patch.object(DatasetManager, 'table_exists_or_error')
    def test_should_use_proper_queries(self, table_exists_or_error_mock, create_full_table_id_mock, collect__mock,
                                       create_table_mock, write_mock, create_bigquery_client_mock):
        create_bigquery_client_mock.return_value = mock.create_autospec(bigquery.Client, project='my_gcp_project', credentials=None, location='EU')
        table_exists_or_error_mock.return_value = True
        internationalports_workflow.run('2019-01-01')
        collect__mock.assert_called_with('''
        INSERT INTO `gcp-project.bigflow_test.ports` (port_name, port_latitude, port_longitude, country, index_number)
        VALUES 
        ('GDYNIA', 54.533333, 18.55, 'POL', '28740'),
        ('GDANSK', 54.35, 18.666667, 'POL', '28710'),
        ('SANKT-PETERBURG', 59.933333, 30.3, 'RUS', '28370'),
        ('TEXAS', 34.8, 31.3, 'USA', '28870');
        ''')

        create_table_mock.assert_any_call('''
    CREATE TABLE IF NOT EXISTS polish_ports (
      port_name STRING,
      port_latitude FLOAT64,
      port_longitude FLOAT64)
''')

        create_table_mock.assert_any_call('''
    CREATE TABLE IF NOT EXISTS ports (
      port_name STRING,
      port_latitude FLOAT64,
      port_longitude FLOAT64,
      country STRING,
      index_number STRING)
''')
        write_mock.assert_called_with('gcp-project.bigflow_test.ports', '''
        SELECT port_name, port_latitude, port_longitude
        FROM `gcp-project.bigflow_test.ports`
        WHERE country = 'POL'
        ''', 'WRITE_TRUNCATE')


from pathlib import Path

from google.cloud.storage import Bucket
from google.oauth2 import credentials
from subprocess import run, CompletedProcess
import requests
from typing import List
from google.cloud import storage

from .commons import run_process, decode_version_number_from_file_name, remove_docker_image_from_local_registry, build_docker_image_tag


def os_call(cmd: List, input: str = None) -> CompletedProcess:
    print('executing cmd:', ' '.join(cmd))
    if input:
        result = run(cmd, check=True, input=input, encoding='ascii')
    else:
        result = run(cmd, check=True)
    print('')
    return result


def load_image_from_tar(image_tar_path: str) -> str:
    for line in run_process(['docker', 'load', '-i', image_tar_path]).split('\n'):
        if 'Loaded image ID:' in line:
            return line.split(' ')[-1].split(':')[-1]
    raise ValueError(f"Can't load image: {image_tar_path}")


def tag_image(image_id, repository, tag):
    os_call(['docker', 'tag', image_id, f'{repository}:{tag}'])


def deploy_docker_image(
        image_tar_path: str,
        docker_repository: str,
        auth_method: str = 'local_account',
        vault_endpoint: str = None,
        vault_secret: str = None):
    build_ver = decode_version_number_from_file_name(Path(image_tar_path))
    image_id = load_image_from_tar(image_tar_path)
    try:
        return _deploy_image_loaded_to_local_registry(build_ver, docker_repository, image_id, auth_method,
                                                      vault_endpoint, vault_secret)
    finally:
        remove_docker_image_from_local_registry(build_docker_image_tag(docker_repository, build_ver))


def _deploy_image_loaded_to_local_registry(build_ver, docker_repository, image_id, auth_method, vault_endpoint,
                                           vault_secret):
    tag_image(image_id, docker_repository, build_ver)
    docker_image = docker_repository + ":" + build_ver
    print(f"Deploying docker image tag={docker_image} auth_method={auth_method}")
    if auth_method == 'local_account':
        os_call(['gcloud', 'auth', 'configure-docker'])
    elif auth_method == 'vault':
        oauthtoken = get_vault_token(vault_endpoint, vault_secret)
        os_call(['docker', 'login', '-u', 'oauth2accesstoken', '--password-stdin', 'https://eu.gcr.io'],
                input=oauthtoken)
    else:
        raise ValueError('unsupported auth_method: ' + auth_method)
    os_call(['docker', 'push', docker_image])
    return docker_image


def deploy_dags_folder(dags_dir: str, dags_bucket: str, project_id: str, clear_dags_folder: bool = False,
                       auth_method: str = 'local_account', vault_endpoint: str = None, vault_secret: str = None,
                       gs_client=None):

    print(f"Deploying DAGs folder, auth_method={auth_method}, clear_dags_folder={clear_dags_folder}, dags_dir={dags_dir}")

    client = gs_client or create_storage_client(auth_method, project_id, vault_endpoint, vault_secret)
    bucket = client.bucket(dags_bucket)

    if clear_dags_folder:
        clear_remote_DAGs_bucket(bucket)

    upload_DAGs_folder(dags_dir, bucket)
    return dags_bucket


def clear_remote_DAGs_bucket(bucket: Bucket):
    i = 0
    for blob in bucket.list_blobs(prefix='dags'):
        if not blob.name in ['dags/', 'dags/airflow_monitoring.py']:
            print(f"deleting file {blob_URI(blob)}")
            blob.delete()
            i += 1

    print(f"{i} files deleted")


def blob_URI(blob):
    return f"gs://{blob.bucket.name}/{blob.name}"


def upload_DAGs_folder(dags_dir: str, bucket: Bucket):
    dags_dir_path = Path(dags_dir)

    def upload_file(local_file_path, target_file_name):
        blob = bucket.blob(target_file_name)
        blob.upload_from_filename(local_file_path, content_type='application/octet-stream')
        print(f"uploading file {local_file_path} to {blob_URI(blob)}")

    i = 0
    for f in dags_dir_path.iterdir():
        if f.is_file():
            upload_file(f.as_posix(), 'dags/' + f.name)
            i += 1

    print(f"{i} files uploaded")


def create_storage_client(auth_method: str, project_id: str, vault_endpoint: str, vault_secret: str):
    if auth_method == 'local_account':
        return storage.Client(project=project_id)
    elif auth_method == 'vault':
        oauthtoken = get_vault_token(vault_endpoint, vault_secret)
        return storage.Client(project=project_id, credentials=credentials.Credentials(oauthtoken))
    else:
        raise ValueError('unsupported auth_method: ' + auth_method)


def get_vault_token(vault_endpoint: str, vault_secret: str):
    if not vault_endpoint:
        raise ValueError('vault_endpoint is required')
    if not vault_secret:
        raise ValueError('vault_secret is required')

    headers = {'X-Vault-Token': vault_secret}
    response = requests.get(vault_endpoint, headers=headers, verify=False)

    if response.status_code != 200:
        print(response.text)
        raise ValueError(
            'Could not get vault token, response code: {}'.format(
                response.status_code))

    print(f"get oauth token from {vault_endpoint} status_code={response.status_code}")
    return response.json()['data']['token']



import unittest
import pathlib
import contextlib
import io

from unittest import mock

import google.cloud.dataproc_v1
from google.cloud.dataproc_v1.services import job_controller
import google.cloud.storage


import bigflow.dataproc
import bigflow.resources


_someobject_tags = set()

def _some_global_func():
    _someobject_tags.add('global_func')


class _SomeObject:

    def __call__(self):
        _someobject_tags.add('__call__')

    def method(self):
        _someobject_tags.add('method')

    @staticmethod
    def static_method():
        _someobject_tags.add('static_method')

    @classmethod
    def class_method(cls):
        _someobject_tags.add('class_method')



class PySparkJobTest(unittest.TestCase):

    def tearDown(self):
        try:
            (pathlib.Path(__file__).parent / "example_project/setup.py").unlink()
        except FileNotFoundError:
            pass

    def test_generates_driver_for_callable(self):

        for callable, expected_tag in [
            (_some_global_func, 'global_func'),
            (_SomeObject(), '__call__'),
            (_SomeObject().static_method, 'static_method'),
            (_SomeObject().class_method, 'class_method'),
            (_SomeObject().method, 'method'),
        ]:
            # given
            _someobject_tags.clear()

            # when
            script = bigflow.dataproc.generate_driver_script(callable)
            exec(script)

            # then 
            # callable was serialized/deserialized/called
            self.assertIn(expected_tag, _someobject_tags)

    
    def test_uploads_driver_to_gcs(self):
        # given
        driver_callable = _some_global_func
        driver_path = 'some-driver.py'

        gsbucket = mock.Mock()
        gsbucket.blob = mock.Mock()

        # when
        bigflow.dataproc.generate_and_upload_driver(driver_callable, gsbucket, driver_path)

        # then
        gsbucket.blob.assert_called_with(driver_path)
        gsbucket.blob.return_value.upload_from_string.assert_called_once_with(
            data=mock.ANY,  # don't werify
            content_type='application/octet-stream',
        )

    def test_build_python_egg_file_for_project(self):
        
        # given
        setup_py = pathlib.Path(__file__).parent / "example_project/setup.py"
        bigflow.resources.find_or_create_setup_for_main_project_package(
            'main_package',
            pathlib.Path(__file__).parent / "example_project/main_project/job",
        )

        # when
        egg = bigflow.dataproc._build_project_egg(setup_py)

        # then
        egg_path = pathlib.Path(egg)
        self.assertTrue(egg_path.exists())

        # cleanup
        egg_path.unlink()
        setup_py.unlink()

    def test_should_create_new_cluster_for_job(self):

        # given
        cluster_name = "test-cluster"
        project_id = "some-project"
        region = "us-west1"

        cluster_client = mock.Mock()

        # when
        bigflow.dataproc._create_cluster(
            cluster_client, 
            project_id,
            region,
            cluster_name,
            requirements=["pkg1==1", "pkg2==2"],
            worker_machine_type='standart-n1',
            worker_num_instances=4,
        )

        # then
        cluster_client.create_cluster.assert_called_once_with(
            project_id=project_id, region=region, cluster=mock.ANY)

        (_, call_kwargs) = cluster_client.create_cluster.call_args
        cluster_data = call_kwargs['cluster']
        cluster_client.create_cluster.return_value.result.assert_called_once()

        self.assertEqual(cluster_data['cluster_name'], cluster_name)
        self.assertEqual(cluster_data['config']['worker_config']['num_instances'], 4)
        self.assertEqual(cluster_data['config']['worker_config']['machine_type_uri'], 'standart-n1')

        self.assertIn(
            ('PIP_PACKAGES', "pkg1==1 pkg2==2"),
            cluster_data['config']['gce_cluster_config']['metadata'],
        )

    def test_should_detect_project_root(self):

        # given
        from test.example_project.main_package.job import create_pyspark_job
        pyspark_job = create_pyspark_job()

        # when
        self.assertIsNone(pyspark_job.setup_file)
        pyspark_job._ensure_has_setup_file()

        # then
        self.assertEqual(
            pyspark_job.setup_file,
            pathlib.Path(__file__).parent / "example_project/setup.py",
        )

    def test_should_send_correct_job_request_to_dataproc(self):

        # given
        cluster_name = "test-cluster"
        project_id = "some-project"
        region = "us-west1"
        bucket_id = "the-bucket"
        driver_path = "some-path/driver.py"
        egg_path = "another-path/file.egg"

        dataproc_job_client = mock.Mock()
        dataproc_job_client.submit_job.return_value.reference.job_id = "the-job"

        # when
        job_id = bigflow.dataproc._submit_single_pyspark_job(
            dataproc_job_client=dataproc_job_client,
            project_id=project_id,
            region=region,
            cluster_name=cluster_name,
            bucket_id=bucket_id,
            driver_path=driver_path,
            egg_path=egg_path,
            jar_file_uris=(),
            properties=(),
        )

        # then
        dataproc_job_client.submit_job.assert_called_once_with(
            project_id=project_id, region=region, job=mock.ANY)
            
        (_, call_kwargs) = dataproc_job_client.submit_job.call_args
        job = call_kwargs['job']

        self.assertEqual(job['placement']['cluster_name'], cluster_name)
        self.assertEqual(job_id, "the-job")

    @mock.patch('google.cloud.storage.Client')
    @mock.patch('google.cloud.dataproc_v1.ClusterControllerClient')
    @mock.patch('google.cloud.dataproc_v1.JobControllerClient')
    def test_run_pyspark_job(
        self,
        job_controller_client_cls,
        cluster_controller_client_cls,
        storage_client_cls,
    ):
        # given
        from test.example_project.main_package.job import create_pyspark_job
        pyspark_job = create_pyspark_job()

        storage = storage_client_cls.return_value
        cluster_controller = cluster_controller_client_cls.return_value
        job_controller = job_controller_client_cls.return_value

        bucket = storage.get_bucket.return_value

        submit_job_result = mock.Mock()
        submit_job_result.reference.job_id = "the-job"
        job_controller.submit_job = mock.Mock(return_value=submit_job_result)

        job_state = mock.Mock()
        job_state.status.state = google.cloud.dataproc_v1.JobStatus.State.DONE
        job_state.driver_output_resource_uri = "gs://job-output"
        job_controller.get_job = mock.Mock(return_value=job_state)

        # when
        with self.assertLogs() as logs:
            pyspark_job.run("2020-02-02")

        # then
        cluster_controller.create_cluster.assert_called_once()
        cluster_controller.delete_cluster.assert_called_once()

        job_controller.submit_job.assert_called_once_with(
            project_id=pyspark_job.gcp_project_id,
            region=pyspark_job.gcp_region,
            job=mock.ANY,
        )

        self.assertTrue(any("JOB OUTPUT:" in line for line in logs.output))



import os
from unittest import TestCase, mock
from pathlib import Path
from bigflow.resources import *

TEST_PROJECT_PATH = Path(__file__).parent / 'example_project'


class FindAllResourcesTestCase(TestCase):
    def test_should_find_all_resource_file_paths(self):
        # when
        resources = list(find_all_resources(TEST_PROJECT_PATH / 'resources'))

        # then
        self.assertCountEqual(resources, [
            'resources/requirements.txt',
            'resources/requirements_base.txt',
        ])


class ReadRequirementsTestCase(TestCase):
    def test_should_return_all_requirements_from_the_hierarchy(self):
        # when
        requirements = read_requirements(TEST_PROJECT_PATH / 'resources' / 'requirements.txt')

        # then
        self.assertEqual(requirements, [
            'freezegun==0.3.14',
            'schedule',
            'datetime_truncate==1.1.0',
        ])


class FindFileTestCase(TestCase):
    def test_should_find_file_going_up_through_hierarchy(self):
        # when
        result_path = find_file('LICENSE', Path(__file__))

        # then
        self.assertEqual(result_path, Path(__file__).parent.parent / 'LICENSE')

    def test_should_raise_error_when_file_not_found(self):
        # then
        with self.assertRaises(ValueError) as e:
            # when
            find_file('LICENSE', Path(__file__), max_depth=1)


class GetResourceAbsolutePathTestCase(TestCase):
    def test_should_find_resource_path(self):
        # when
        result_path = get_resource_absolute_path('test_resource', Path(__file__))

        # then
        self.assertEqual(result_path, Path(__file__).parent / 'resources' / 'test_resource')

    def test_should_raise_error_when_resource_not_found(self):
        # then
        with self.assertRaises(ValueError) as e:
            # when
            get_resource_absolute_path('unknown_resource', Path(__file__))


class FindSetupTestCase(TestCase):
    def test_should_find_setup(self):
        # when
        setup_path = find_setup(Path(__file__))

        # then
        self.assertEqual(setup_path, Path(__file__).parent.parent / 'setup.py')

    @mock.patch('bigflow.resources.find_file')
    def test_should_retry_n_times(self, find_file_mock):
        # given
        find_file_mock.side_effect = self.raise_value_error

        # then
        with self.assertRaises(ValueError) as e:
            # when
            find_setup(Path(__file__), retries_left=2, sleep_time=0.01)

        # and
        self.assertEqual(find_file_mock.call_count, 3)

    def raise_value_error(self, *args, **kwargs):
        raise ValueError('Setup file not found')


class CreateFileIfNotExistsTestCase(TestCase):
    def setUp(self):
        self.example_file = Path(__file__).parent / 'resources' / 'example_file'
        if self.example_file.exists():
              os.remove(self.example_file)

    def test_should_create_existent_or_non_existent_file(self):
        # when
        create_file_if_not_exists(self.example_file, 'test file')

        # then
        self.assertTrue(self.example_file.read_text(), 'test file')

        # when trying to create existing
        create_file_if_not_exists(self.example_file, 'test file')

        # then
        self.assertTrue(self.example_file.read_text(), 'test file')


class CreateSetupBodyTestCase(TestCase):
    def test_should_create_setup_body(self):
        # when
        body = create_setup_body('example_project')

        # then
        self.assertEqual(body, f'''
import setuptools

setuptools.setup(
        name='example_project',
        version='0.1.0',
        packages=setuptools.find_packages(
                exclude=tuple(
                        p for p in setuptools.find_packages()
                        if not p.startswith('example_project'))))
''')


class FindOrCreateSetupForMainProjectPackageTestCase(TestCase):
    def setUp(self):
        self.setup_py = Path(__file__).parent / 'example_project' / 'setup.py'
        if self.setup_py.exists():
              os.remove(self.setup_py)

    def test_should_find_setup_if_exists_or_create_it_next_to_main_package(self):
        # when does not exist
        path = find_or_create_setup_for_main_project_package(
            'main_package', Path(__file__).parent / 'example_project' / 'main_package' / 'job.py')

        # then
        self.assertTrue(self.setup_py.exists())
        self.assertEqual(path, self.setup_py)

        # when already exists
        path = find_or_create_setup_for_main_project_package(
            'main_package', Path(__file__).parent / 'example_project' / 'main_package' / 'job.py')

        # then
        self.assertEqual(path, self.setup_py)


class FindOrCreateSetupForDefaultProjectPackageTestCase(TestCase):
    def setUp(self):
        self.setup_py = Path(__file__).parent.parent / 'setup.py'

    def test_should_find_existing_setup_py_with_default_args(self):
        path = find_or_create_setup_for_main_project_package()

        self.assertEqual(path, self.setup_py)



import os
from pathlib import Path

import bigflow

from bigflow.scaffold.scaffold_templates import beam_workflow_template, beam_processing_template, \
    beam_pipeline_template, project_setup_template, basic_deployment_config_template, \
    advanced_deployment_config_template, docker_template, basic_beam_config_template, requirements_template, \
    readme_template, advanced_beam_config_template, gitignore_template, test_wordcount_workflow_template, bq_workflow_template


def start_project(config):
    templates = format_templates(config)
    create_dirs_and_files(config, templates)


def format_templates(config):
    beam_config_template = basic_beam_config_template.format(project_id=config['projects_id'][0], project_name=config['project_name'])
    test_templates = {
        '__init__.py': '',
        'test_wordcount_workflow.py': test_wordcount_workflow_template.format(project_name=config['project_name'])}
        
    resources_templates = {
        'requirements.txt': requirements_template.format(
            bigflow_version=bigflow.__version__,
        ),
    }
    deployment_config_template = basic_deployment_config_template.format(project_id=config['projects_id'][0], dags_bucket=config['composers_bucket'][0])

    if not config['is_basic']:
        for i in range(1, len(config['projects_id'])):
            deployment_config_template = deployment_config_template.strip()
            deployment_config_template += advanced_deployment_config_template.format(env=config['envs'][i], project_id=config['projects_id'][i], dags_bucket=config['composers_bucket'][i])

            beam_config_template = beam_config_template.strip()
            beam_config_template += advanced_beam_config_template.format(env=config['envs'][i], project_id=config['projects_id'][i], dags_bucket=config['composers_bucket'][i])

    beam_config_template = beam_config_template.strip()
    beam_config_template += '.resolve()'

    main_templates = {
        'project_setup.py': project_setup_template.format(project_name=config['project_name']),
        'deployment_config.py': deployment_config_template,
        'Dockerfile': docker_template,
        'README.md': readme_template.format(project_name=config['project_name'])}
    beam_templates = {
        'config.py': beam_config_template,
        'workflow.py': beam_workflow_template,
        'processing.py': beam_processing_template,
        'pipeline.py': beam_pipeline_template,
        '__init__.py': ''}
    bq_templates = {
        '__init__.py': '',
        'workflow.py': bq_workflow_template % {'project_id': config['projects_id'][0]}}
    return {
        'beam_templates': beam_templates,
        'bq_templates': bq_templates,
        'test_templates': test_templates,
        'resources_templates': resources_templates,
        'main_templates': main_templates}


def create_dirs_and_files(config, templates):
    project_dir = config['project_name'] + '_project'
    project_path = Path(project_dir).resolve()
    workflows_path = Path(project_dir) / config['project_name']
    word_count_path = workflows_path / 'wordcount'
    internationalports_path = workflows_path / 'internationalports'
    resources_path = Path(project_dir + '/resources')
    test_path = Path(project_dir + '/test')

    os.mkdir(project_path)
    os.mkdir(workflows_path)
    os.mkdir(word_count_path)
    os.mkdir(internationalports_path)
    os.mkdir(resources_path)
    os.mkdir(test_path)

    create_module_files(templates['main_templates'], project_path.resolve())
    create_module_files(templates['beam_templates'], word_count_path.resolve())
    create_module_files(templates['bq_templates'], internationalports_path.resolve())
    create_module_files(templates['resources_templates'], resources_path.resolve())
    create_module_files(templates['test_templates'], test_path.resolve())
    create_module_files({'__init__.py': ''}, workflows_path.resolve())
    create_module_files({'.gitignore': gitignore_template}, project_path.resolve())


def create_module_files(templates, path):
    for filename, template in templates.items():
        with open(os.path.join(path, filename), 'w+') as f:
            f.write(template)



import os
from pathlib import Path
import shutil
import distutils.cmd
import setuptools
import unittest
from setuptools import setup

import xmlrunner

from .cli import walk_workflows, import_deployment_config, _valid_datetime, SETUP_VALIDATION_MESSAGE
from .dagbuilder import generate_dag_file
from .resources import read_requirements, find_all_resources
from .commons import resolve, now, run_process, remove_docker_image_from_local_registry, get_docker_image_id, build_docker_image_tag
from .version import get_version


__all__ = [
    'project_setup',
    'auto_configuration',
    'default_project_setup'
]


def run_tests(build_dir: Path, test_package: Path):
    runner = xmlrunner.XMLTestRunner(output=resolve(build_dir / 'junit-reports'))
    suite = unittest.TestLoader().discover(start_dir=resolve(test_package))
    result = runner.run(suite)
    if result.errors:
        raise ValueError('Test suite failed.')


def export_docker_image_to_file(tag: str, target_dir: Path, version: str):
    image_target_path = target_dir / f'image-{version}.tar'
    print(f'Exporting the image to file: {image_target_path}' )
    run_process(f"docker image save {get_docker_image_id(tag)} -o {resolve(image_target_path)}")


def build_docker_image(project_dir: Path, tag: str):
    print('Building a Docker image. It might take a while.')
    run_process(f'docker build {resolve(project_dir)} --tag {tag}')


def build_dags(
        root_package,
        project_dir,
        docker_repository,
        start_time,
        version,
        specific_workflow=None):
    for workflow in walk_workflows(root_package):
        if specific_workflow is not None and specific_workflow != workflow.workflow_id:
            continue
        print(f'Generating DAG file for {workflow.workflow_id}')
        generate_dag_file(
            resolve(project_dir),
            docker_repository,
            workflow,
            start_time,
            version,
            resolve(root_package).split(os.sep)[-1])


def build_image(
        docker_repository: str,
        version: str,
        project_dir: Path,
        image_dir: Path,
        deployment_config: Path):
    os.mkdir(image_dir)
    tag = build_docker_image_tag(docker_repository, version)
    build_docker_image(project_dir, tag)
    try:
        export_docker_image_to_file(tag, image_dir, version)
        shutil.copyfile(deployment_config, image_dir / resolve(deployment_config).split(os.sep)[-1])
    finally:
        remove_docker_image_from_local_registry(tag)


def build_command(
        root_package: Path,
        project_dir: Path,
        build_dir: Path,
        test_package: Path,
        dags_dir: Path,
        dist_dir: Path,
        image_dir: Path,
        eggs_dir: Path,
        deployment_config: Path,
        docker_repository: str,
        version: str):

    class BuildCommand(distutils.cmd.Command):
        description = 'BigFlow project build.'
        user_options = [
            ('build-dags', None, 'Builds the DAG files.'),
            ('build-package', None, 'Builds the whl package.'),
            ('build-image', None, 'Builds the Docker image.'),
            ('start-time=', None, 'DAGs start time -- given in local timezone, for example: 2020-06-27 15:00:00'),
            ('workflow=', None, 'The workflow that you want to build DAG for.'),
            ('validate-project-setup', None, 'If used, echoes a message that can be used by the CLI to determine if the setup is working.'),
        ]

        def initialize_options(self) -> None:
            self.start_time = now()
            self.build_dags = False
            self.build_package = False
            self.build_image = False
            self.workflow = None
            self.validate_project_setup = False

        def should_run_whole_build(self):
            return not self.build_dags and not self.build_package and not self.build_image

        def finalize_options(self) -> None:
            pass

        def run(self) -> None:
            if self.validate_project_setup:
                print(SETUP_VALIDATION_MESSAGE)
                return

            _valid_datetime(self.start_time)
            if self.build_package or self.should_run_whole_build():
                print('Building the pip package')
                clear_package_leftovers(dist_dir, eggs_dir, build_dir)
                run_tests(build_dir, test_package)
                self.run_command('bdist_wheel')

            if self.build_dags or self.should_run_whole_build():
                print('Building the dags')
                clear_dags_leftovers(dags_dir)
                build_dags(
                    root_package,
                    project_dir,
                    docker_repository,
                    self.start_time,
                    version,
                    self.workflow)

            if self.build_image or self.should_run_whole_build():
                print('Building the image')
                clear_image_leftovers(image_dir)
                build_image(
                    docker_repository,
                    version,
                    project_dir,
                    image_dir,
                    deployment_config)

    return BuildCommand


def clear_image_leftovers(image_dir: Path):
    print(f'Removing: {str(image_dir.absolute())}')
    shutil.rmtree(image_dir, ignore_errors=True)


def clear_dags_leftovers(dags_dir: Path):
    print(f'Removing: {str(dags_dir.absolute())}')
    shutil.rmtree(dags_dir, ignore_errors=True)


def clear_package_leftovers(dist_dir: Path, eggs_dir: Path, build_dir: Path):
    for to_delete in [build_dir, dist_dir, eggs_dir]:
        print(f'Removing: {str(to_delete.absolute())}')
        shutil.rmtree(to_delete, ignore_errors=True)


def get_docker_repository_from_deployment_config(deployment_config_file: Path):
    try:
        config = import_deployment_config(resolve(deployment_config_file), 'docker_repository')
    except ValueError:
        raise ValueError(f"Can't find the specified deployment configuration: {resolve(deployment_config_file)}")
    docker_repository = config.resolve_property('docker_repository', None)

    if docker_repository is None:
        raise ValueError(f"Can't find the 'docker_repository' property in the specified config file: {resolve(deployment_config_file)}")
    return docker_repository


def secure_get_version():
    try:
        return get_version()
    except Exception as e:
        print(e)
        raise ValueError("Can't get the current package version. To use the automatic versioning, "
                         "you need to use git inside your project directory.")


def auto_configuration(project_name: str, project_dir: Path = Path('.').parent):
    '''
    Auto configuration for the standard BigFlow project structure (that you can generate through the CLI).
    The 'project_name' parameter should be a valid python package name.

    Example:
    project_setup(**auto_configuration('my_super_project'))
    '''
    deployment_config_file = project_dir / 'deployment_config.py'

    return {
        'project_name': project_name,
        'docker_repository': get_docker_repository_from_deployment_config(deployment_config_file),
        'root_package': project_dir / project_name,
        'project_dir': project_dir,
        'build_dir': project_dir / 'build',
        'test_package': project_dir / 'test',
        'dags_dir': project_dir / '.dags',
        'dist_dir': project_dir / 'dist',
        'image_dir': project_dir / 'image',
        'eggs_dir': project_dir / f'{project_name}.egg-info',
        'deployment_config_file': deployment_config_file,
        'version': secure_get_version(),
        'resources_dir': project_dir / 'resources',
        'project_requirements_file': project_dir / 'resources' / 'requirements.txt'
    }


def parameter_not_null_or_error(parameter_name: str, parameter_value):
    if parameter_value is None:
        raise ValueError(f"Parameter {parameter_name} can't be None.")


def project_setup(
        project_name: str,
        docker_repository: str,
        root_package: Path,
        project_dir: Path,
        build_dir: Path,
        test_package: Path,
        dags_dir: Path,
        dist_dir: Path,
        image_dir: Path,
        eggs_dir: Path,
        deployment_config_file: Path,
        version: str,
        resources_dir: Path,
        project_requirements_file: Path):
    '''
    This function produces arguments for setuptools.setup. The produced setup provides commands that allow you to build
    whl package, docker image and DAGs. Paired with auto_configuration function, it provides fully automated build
    tool (accessed by CLI) for your project (that you can generate using CLI).

    Example:
    from setuptools import setup
    from bigflow.build import project_setup, auto_configuration

    setup(project_setup(**auto_configuration('my_super_project')))
    '''
    params_to_check = [
        ('project_name', project_name),
        ('docker_repository', docker_repository),
        ('root_package', root_package),
        ('project_dir', project_dir),
        ('build_dir', build_dir),
        ('test_package', test_package),
        ('dags_dir', dags_dir),
        ('dist_dir', dist_dir),
        ('image_dir', image_dir),
        ('eggs_dir', eggs_dir),
        ('deployment_config_file', deployment_config_file),
        ('version', version),
        ('resources_dir', resources_dir),
        ('project_requirements_file', project_requirements_file),
    ]
    for parameter_name, parameter_value in params_to_check:
        parameter_not_null_or_error(parameter_name, parameter_value)

    return {
        'name': project_name,
        'version': version,
        'packages': setuptools.find_packages(exclude=['test']),
        'install_requires': read_requirements(project_requirements_file),
        'data_files': [
            ('resources', list(find_all_resources(resources_dir)))
        ],
        'cmdclass': {
            'build_project': build_command(
                root_package,
                project_dir,
                build_dir,
                test_package,
                dags_dir,
                dist_dir,
                image_dir,
                eggs_dir,
                deployment_config_file,
                docker_repository,
                version)
        }
    }


def default_project_setup(project_name, project_dir: Path = Path('.').parent):
    return setup(**project_setup(**auto_configuration(project_name, project_dir=project_dir)))


import logging
import sys
import uuid
import traceback

from textwrap import dedent

import bigflow

from google.cloud import logging_v2
from google.cloud.logging_v2.gapic.enums import LogSeverity
from urllib.parse import quote_plus

try:
    from typing import TypedDict
except ImportError:
    TypedDict = dict


logger = logging.getLogger(__name__)


class GCPLoggerHandler(logging.Handler):

    def __init__(self, project_id, log_name, labels):
        super().__init__()

        self.client = self.create_logging_client()
        self.project_id = project_id
        self.log_name = log_name
        self.labels = labels

        self._log_entry_prototype = logging_v2.types.LogEntry(
            log_name=f"projects/{self.project_id}/logs/{self.log_name}",
            labels=dict(labels),
            resource={
                "type": "global",
                "labels": {
                    "project_id": str(self.project_id),
                },
            },
        )

    def create_logging_client(self):
        return logging_v2.LoggingServiceV2Client()

    def emit(self, record: logging.LogRecord):

        # CloudLogging list of supported log levels is a superset of python logging level names
        cl_log_level = record.levelname

        json = {
            'message': record.getMessage(),
            'name': record.name,
            'filename': record.filename,
            'module': record.module,
            'lineno': record.lineno,
        }

        # mimic caching behaviour of `logging.Formatter.format`
        # render/cache exception info in the same way as default `logging.Formatter`
        if record.exc_info and not record.exc_text:
            record.exc_text = self._format_exception(record.exc_info)
        if record.exc_text:
            json['exc_text'] = str(record.exc_text)

        if record.stack_info:
            json['stack_info'] = str(record.stack_info)

        self.write_log_entries(json, cl_log_level)

    def _format_exception(self, exc_info):
        etype, value, tb = exc_info
        return "\n".join(traceback.format_exception(etype, value, tb))

    def write_log_entries(self, json, severity):
        entry = logging_v2.types.LogEntry()
        entry.CopyFrom(self._log_entry_prototype)
        entry.json_payload.update(json)
        entry.severity = LogSeverity[severity]
        self.client.write_log_entries([entry])


def prepare_gcp_logs_link(query: str):
    return f"https://console.cloud.google.com/logs/query;query={quote_plus(query)}"


def _uncaught_exception_handler(logger):
    def handler(exception_type, value, traceback):
        logger.error(f'Uncaught exception: {value}', exc_info=(exception_type, value, traceback))
    return handler


_LOGGING_CONFIGURED = False


# TODO: move to some shared config module?
class LogConfigDict(TypedDict):
    gcp_project_id: str
    log_name: str
    verbose: bool


def _generate_cl_log_view_query(params: dict):
    query = "".join(
        '{}"{}"\n'.format(k, v)
        for k, v in params.items()
    )
    return query


def workflow_logs_link_for_cli(log_config, workflow_id):
    log_name = log_config.get('log_name', workflow_id)
    full_log_name = f"projects/{log_config['gcp_project_id']}/logs/{log_name}"
    return prepare_gcp_logs_link(
        _generate_cl_log_view_query({'logName=': full_log_name, 'labels.workflow_id=': workflow_id}))


def infrastructure_logs_link_for_cli(projects_config):
    links = {}
    for project_id, workflow_id in projects_config:
        links[project_id] = get_infrastrucutre_bigflow_project_logs(project_id, workflow_id)
    return links


def print_log_links_message(workflows_links, infra_links):
    workflows_links = '\n'.join(f'{workflow}: {link}' for workflow, link in workflows_links.items())
    infra_links = '\n'.join(f'{workflow}: {link}' for workflow, link in infra_links.items())
    print(dedent(f"""
*************************LOGS LINK*************************
Infrastructure logs: 
{infra_links}
Workflow logs: 
{workflows_links}
***********************************************************"""))


def get_infrastrucutre_bigflow_project_logs(project_id, workflow_id):
    pod_errors = _generate_cl_log_view_query({"severity>=": "INFO"}) + _generate_cl_log_view_query(
        {"resource.type=": "k8s_pod"}) + '"Error:"'
    container_errors = _generate_cl_log_view_query({"severity>=": "INFO"}) + _generate_cl_log_view_query(
        {"resource.type=": "k8s_container", "resource.labels.container_name=": "base"})
    dataflow_errors = _generate_cl_log_view_query({"resource.type=": "dataflow_step",
                                                   "log_name=": f"projects/{project_id}/logs/dataflow.googleapis.com%2Fjob-message",
                                                   "labels.workflow_id=": workflow_id}) + _generate_cl_log_view_query(
        {"severity>=": "INFO"})

    result = []
    for entry in [pod_errors, container_errors, dataflow_errors]:
        result.append(f'({entry})')

    condition = '\nOR\n'.join(result)
    return prepare_gcp_logs_link(condition)


def init_workflow_logging(workflow: 'bigflow.Workflow'):
    if workflow.log_config:
        init_logging(workflow.log_config, workflow.workflow_id)
    else:
        print("Log config is not provided for the Workflow")


def init_logging(config: LogConfigDict, workflow_id: str):
    global _LOGGING_CONFIGURED
    if _LOGGING_CONFIGURED:
        import warnings
        warnings.warn(UserWarning("bigflow.log is already configured - skip"))
        return
    _LOGGING_CONFIGURED = True

    gcp_project_id = config['gcp_project_id']
    log_name = config.get('log_name', workflow_id)
    run_uuid = str(uuid.uuid4())

    labels = {
        'workflow_id': workflow_id,
        'run_uuid': run_uuid,
    }

    root = logging.getLogger()
    if not root.handlers:
        root.setLevel(logging.INFO)
        h1 = logging.StreamHandler(sys.stdout)
        h1.setLevel(logging.INFO)
        h1.addFilter(lambda record: record.levelno <= logging.WARNING)
        h2 = logging.StreamHandler(sys.stderr)
        h2.setLevel(logging.ERROR)
        root.addHandler(h1)
        root.addHandler(h2)

    full_log_name = f"projects/{gcp_project_id}/logs/{log_name}"
    infrastructure_logs = get_infrastrucutre_bigflow_project_logs(gcp_project_id, workflow_id)
    workflow_logs_link = prepare_gcp_logs_link(
        _generate_cl_log_view_query({'logName=': full_log_name, 'labels.workflow_id=': workflow_id}))
    this_execution_logs_link = prepare_gcp_logs_link(
        _generate_cl_log_view_query({'logName=': full_log_name, 'labels.run_uuid=': run_uuid}))

    logger.info(dedent(f"""
           *************************LOGS LINK*************************
           Infrastructure logs:{infrastructure_logs}
           Workflow logs (all runs): {workflow_logs_link}
           Only this run logs: {this_execution_logs_link}
           ***********************************************************"""))
    gcp_logger_handler = GCPLoggerHandler(gcp_project_id, log_name, labels)
    gcp_logger_handler.setLevel(logging.INFO)
    # TODO: add formatter?
    root.addHandler(gcp_logger_handler)

    sys.excepthook = _uncaught_exception_handler(logging.getLogger('uncaught_exception'))



readme_template = '''## Run project

### Requirements & preparations
[Installing BigFlow](https://github.com/allegro/bigflow/blob/master/README.md#installing-bigflow)

[Preparations](https://github.com/allegro/bigflow/blob/master/docs/scaffold.md#preparation)

### Running locally
  1. To run bigquery workflow type in project directory. 
  
    bf run --workflow internationalports
  1. To run apache beam workflow on Dataflow type in project directory. 
  
    bf run --workflow wordcount
  
### Building & deploying
  [Build](https://github.com/allegro/bigflow/blob/master/docs/cli.md#building-airflow-dags) the project:
    
    bigflow build
  
  [Deploy](https://github.com/allegro/bigflow/blob/master/docs/cli.md#deploying-to-gcp) the project to Composer:
    
    bigflow deploy
  
  [Run](https://github.com/allegro/bigflow/blob/master/docs/cli.md#running-jobs-and-workflows) a workflow on given env:
    
    bigflow run ...
  
  
### Results
  1. You can find results of `internationalports` workflow in your BigQuery under `bigflow_test` dataset. The results
  of `wordcount` you can find in your `{project_name}` bucket used by dataflow in `beam_runner/temp` directory.
'''

docker_template = '''FROM python:3.7
COPY ./dist /dist
RUN apt-get -y update && apt-get install -y libzbar-dev libc-dev musl-dev
RUN for i in /dist/*.whl; do pip install $i; done
'''

basic_deployment_config_template = '''from bigflow.configuration import Config

deployment_config = Config(
    name='dev',
    properties={{
       'docker_repository': 'test_repository',
       'gcp_project_id': '{project_id}',
       'dags_bucket': '{dags_bucket}'}})
'''

advanced_deployment_config_template = '''.add_configuration(
                            name='{env}',
                            properties={{
                                'gcp_project_id': '{project_id}',
                                'dags_bucket': '{dags_bucket}',
                            }})
'''
requirements_template = '''bigflow[bigquery,log]=={bigflow_version}
apache-beam[gcp]==2.23.0
'''

project_setup_template = '''from bigflow.build import default_project_setup

PROJECT_NAME = '{project_name}'

if __name__ == '__main__':
    default_project_setup(PROJECT_NAME)
'''

beam_pipeline_template = '''import uuid
import logging
from pathlib import Path

import apache_beam as beam
from apache_beam.options.pipeline_options import SetupOptions, StandardOptions, WorkerOptions, GoogleCloudOptions, PipelineOptions


logger = logging.getLogger(__name__)


def dataflow_pipeline(gcp_project_id, staging_location, temp_location, region, machine_type, project_name):
    from bigflow.resources import find_or_create_setup_for_main_project_package, resolve, get_resource_absolute_path
    options = PipelineOptions()

    google_cloud_options = options.view_as(GoogleCloudOptions)
    google_cloud_options.project = gcp_project_id
    google_cloud_options.job_name = f'beam-wordcount-{uuid.uuid4()}'
    google_cloud_options.staging_location = f"gs://{staging_location}"
    google_cloud_options.temp_location = f"gs://{temp_location}"
    google_cloud_options.region = region

    options.view_as(WorkerOptions).machine_type = machine_type
    options.view_as(WorkerOptions).max_num_workers = 2
    options.view_as(WorkerOptions).autoscaling_algorithm = 'THROUGHPUT_BASED'
    options.view_as(StandardOptions).runner = 'DataflowRunner'

    options.view_as(SetupOptions).setup_file = resolve(find_or_create_setup_for_main_project_package(project_name, Path(__file__)))
    options.view_as(SetupOptions).requirements_file = resolve(get_resource_absolute_path('requirements.txt', Path(__file__)))

    logger.info("Run beam pipeline with options %s", options)
    return beam.Pipeline(options=options)
'''
beam_processing_template = '''import logging

import apache_beam as beam


logger = logging.getLogger(__name__)


def count_words(p, target_method):
    logger.debug("Count words with target method %s", target_method)
    return (p | beam.Create(['a', 'b', 'c', 'd', 'a', 'b', 'c', 'd'])
            | 'PairWithOne' >> beam.Map(lambda x: (x, 1))
            | 'GroupAndSum' >> beam.CombinePerKey(sum)
            | 'save' >> target_method)
'''
beam_workflow_template = '''import logging

from apache_beam.io import WriteToText

import bigflow

from .pipeline import dataflow_pipeline
from .config import workflow_config
from .processing import count_words


logger = logging.getLogger(__name__)


class WordcountJob(bigflow.Job):

    def __init__(self, id, gcp_project_id, staging_location, temp_location, region, machine_type, project_name):
        logger.debug("Init workflow %s", id)

        self.id = id
        self.retry_count = 20
        self.retry_pause_sec = 100

        self.gcp_project_id = gcp_project_id
        self.staging_location = staging_location
        self.temp_location = temp_location
        self.region = region
        self.machine_type = machine_type
        self.project_name = project_name

    def execute(self, context: bigflow.JobContext):
        logger.info("Run workflow %s at %s", self.id, context.runtime)
        p = dataflow_pipeline(self.gcp_project_id, self.staging_location, self.temp_location, self.region,
                              self.machine_type, self.project_name)
        count_words(p, WriteToText("gs://{}/beam_wordcount".format(self.temp_location)))
        p.run().wait_until_finish()


simple_workflow = bigflow.Workflow(
    workflow_id="wordcount",
    log_config={
        'gcp_project_id': workflow_config['gcp_project_id'],
        'log_level': 'INFO',
    },
    definition=[WordcountJob(
        'test_workflow',
        gcp_project_id=workflow_config['gcp_project_id'],
        staging_location=workflow_config['staging_location'],
        temp_location=workflow_config['temp_location'],
        region=workflow_config['region'],
        machine_type=workflow_config['machine_type'],
        project_name=workflow_config['project_name']
    )])
'''
basic_beam_config_template = '''from bigflow.configuration import Config

workflow_config = Config(name='dev',
    properties={{
        'project_name': '{project_name}',
        'gcp_project_id': '{project_id}',
        'staging_location': '{project_name}/beam_runner/staging',
        'temp_location': '{project_name}/beam_runner/temp',
        'region': 'europe-west1',
        'machine_type': 'n1-standard-1'}})
'''

advanced_beam_config_template = '''.add_configuration(name='{env}',
                           properties={{
                               'gcp_project_id': '{project_id}',
                               'dags_bucket': '{dags_bucket}'}})
'''

bq_workflow_template = """
from bigflow import Workflow
from bigflow.bigquery import DatasetConfig

dataset_config = DatasetConfig(
    env='dev',
    project_id='%(project_id)s',
    dataset_name='internationalports',
    internal_tables=['ports', 'polish_ports'],
    external_tables={})

dataset = dataset_config.create_dataset_manager()

create_polish_ports_table = dataset.create_table('''
    CREATE TABLE IF NOT EXISTS polish_ports (
      port_name STRING,
      port_latitude FLOAT64,
      port_longitude FLOAT64)
''')

create_ports_table = dataset.create_table('''
    CREATE TABLE IF NOT EXISTS ports (
      port_name STRING,
      port_latitude FLOAT64,
      port_longitude FLOAT64,
      country STRING,
      index_number STRING)
''')

select_polish_ports = dataset.write_truncate('ports', '''
        SELECT port_name, port_latitude, port_longitude
        FROM `{ports}`
        WHERE country = 'POL'
        ''', partitioned=False)

populate_ports_table = dataset.collect('''
        INSERT INTO `{ports}` (port_name, port_latitude, port_longitude, country, index_number)
        VALUES 
        ('GDYNIA', 54.533333, 18.55, 'POL', '28740'),
        ('GDANSK', 54.35, 18.666667, 'POL', '28710'),
        ('SANKT-PETERBURG', 59.933333, 30.3, 'RUS', '28370'),
        ('TEXAS', 34.8, 31.3, 'USA', '28870');
        ''')


internationalports_workflow = Workflow(
        workflow_id='internationalports',
        definition=[
                create_ports_table.to_job(id='create_ports_table'),
                create_polish_ports_table.to_job(id='create_polish_ports_table'),
                populate_ports_table.to_job(id='populate_ports_table'),
                select_polish_ports.to_job(id='select_polish_ports'),
        ],
        schedule_interval='@once')
"""

test_wordcount_workflow_template = '''from unittest import TestCase

from apache_beam.testing.test_pipeline import TestPipeline
import apache_beam as beam
from {project_name}.wordcount.processing import count_words


class WordCountWorkflowTestCase(TestCase):

    def test_should_return_result(self):
        fake_file = FakeFile()
        with TestPipeline() as p:
            count_words(p, FakeFileSaver(fake_file))
        self.assertEqual(fake_file.data, {{'a': 2, 'b': 2, 'c': 2, 'd': 2}})


class FakeFileSaver(beam.PTransform):
    def __init__(self, file):
        super().__init__()
        self.file = file

    def expand(self, records_to_delete):
        return records_to_delete \\
               | "save to file" >> beam.ParDo(
            SaveFn(self.file))


class SaveFn(beam.DoFn):
    def __init__(self, file):
        super().__init__()
        self.file = file

    def process(self, row, *args, **kwargs):
        self.file.data[row[0]] = row[1]

class FakeFile(object):
    data = {{}}

'''

gitignore_template = '''# Byte-compiled / optimized / DLL files
__pycache__/
*.py[cod]
*$py.class

# C extensions
*.so

# Distribution / packaging
.Python
build/
develop-eggs/
dist/
downloads/
eggs/
.eggs/
lib/
lib64/
parts/
sdist/
var/
wheels/
pip-wheel-metadata/
share/python-wheels/
*.egg-info/
.installed.cfg
*.egg
MANIFEST

# PyInstaller
#  Usually these files are written by a python script from a template
#  before PyInstaller builds the exe, so as to inject date/other infos into it.
*.manifest
*.spec

# Installer logs
pip-log.txt
pip-delete-this-directory.txt

# Unit test / coverage reports
htmlcov/
.tox/
.nox/
.coverage
.coverage.*
.cache
nosetests.xml
coverage.xml
*.cover
.hypothesis/
.pytest_cache/

# Translations
*.mo
*.pot

# Django stuff:
*.log
local_settings.py
db.sqlite3
db.sqlite3-journal

# Flask stuff:
instance/
.webassets-cache

# Scrapy stuff:
.scrapy

# Sphinx documentation
docs/_build/

# PyBuilder
target/

# Jupyter Notebook
.ipynb_checkpoints

# IPython
profile_default/
ipython_config.py

# pyenv
.python-version

# pipenv
#   According to pypa/pipenv#598, it is recommended to include Pipfile.lock in version control.
#   However, in case of collaboration, if having platform-specific dependencies or dependencies
#   having no cross-platform support, pipenv may install dependencies that don't work, or not
#   install all needed dependencies.
#Pipfile.lock

# celery beat schedule file
celerybeat-schedule

# SageMath parsed files
*.sage.py

# Environments
.env
.venv
.venv2
env/
venv/
ENV/
env.bak/
venv.bak/

# Spyder project settings
.spyderproject
.spyproject

# Rope project settings
.ropeproject

# mkdocs documentation
/site

# mypy
.mypy_cache/
.dmypy.json
dmypy.json

# Pyre type checker
.pyre/
.dags
image
.dags/
image/
.idea'''

